import sys
import gfapy
import time

class Logger:
  """
  Output messages to the standard error or a logfile and
  keep track of the progress of long running methods.

  Parameters:
    verbose_level (int) : 0: no logging; >0: the higher, the more logging
      (default: 1); messages output using the log() method can be provided
      with a min_verbose_level, and are output only if that value is equal
      or higher than verbose_level
    channel : where to output (default: sys.stderr), it must provide a
      write() method.
    prefix (str) : output prefix (default: ```#```)

  Returns:
    gfapy.Logger
  """

  class ProgressData:
    """
    Information about the progress of a computation
    """
    def __init__(self, counter, units, partsize, lastpart, total, starttime,
                 strlen):
      self._counter = counter
      self._units = units
      self._partsize = partsize
      self._lastpart = lastpart
      self._total = total
      self._starttime = starttime
      self._strlen = strlen

  def __init__(self, verbose_level = 1, channel = sys.stderr, prefix = "#"):
    self._progress = False
    if not isinstance(verbose_level, int):
      raise gfapy.ArgumentError("verbose_level must be an Integer")
    if not(getattr(channel, "write", None) and callable(channel.write)):
      raise gfapy.TypeError("channel must provide a 'write' method")
    self._channel = channel
    self._pfx = prefix
    self._verbose_level = verbose_level
    self._data = {}

  def log(self, msg, min_verbose_level=1):
    """Output a log message to the logger output channel.

    Parameters:
      msg (str) : message to output
      min_verbose_level (int) : output the message only if the
        verbose level of the logger is at least the specified one
        (default: 1)
    """
    if self._verbose_level >= min_verbose_level:
      self._channel.write("{} {}\n".format(self._pfx, msg))

  def enable_progress(self, part = 0.1):
    """Enable output of progress of long running methods.

    Parameters
      part (float between 0 and 1) : if part = 0, output at every call of
         progress_log(); if 0 < part < 1, output once per part of the total
         progress (e.g. 0.001 = log every 0.1% progress); if part = 1, output
         only total elapsed time at the end of the computation.
    """
    if part < 0 or part > 1:
      raise gfapy.ArgumentError("part must be in range [0..1]")
    self._progress = True
    self._part = part
    if self._verbose_level > 0:
      self._channel.write("{} Progress logging enabled\n".format(self._pfx))

  def disable_progress(self):
    """Disable output of progress of long running methods."""
    self._progress = False
    if self._verbose_level > 0:
      self._channel.write("{} Progress logging disabled\n".format(self._pfx))

  def progress_init(self, symbol, units, total, initmsg = None):
    """Initialize progress logging for a long running computation.

    Parameters:
      symbol (str) : an identifier assigned to the computation
      units (str) : the name of the units of computation, in plural, for the
                    output messages
      total (int) : the total number of units of the computation
      initmsg (str) : an optional message to output at the beginning of the
                      computation
    """
    if not self._progress or total == 0:
      return
    string = "{} 0.0% {} processed".format(self._pfx, units)
    self._data[symbol] = Logger.ProgressData(0, units, int(self._part*total),
                                             1, total, time.time(), len(string))
    if initmsg:
      self._channel.write("{} {}\n".format(self._pfx, initmsg))
    if self._part != 1:
      self._channel.write(string)

  def progress_log(self, symbol, progress=1, **keyargs):
    """Updates progress of a computation.

    A logging message is output or not, depending on the part parameter
    (see the `enable_progress` method).

    Parameters:
      symbol (str) : the identifier assigned to the computation when
                    `progress_init` was called
      progress (int) : how many units of computations were completed
                       in the last interaction (default: 1)
      **keyargs (dict) : additional units of computation to display (keys),
                       together with their current progress value (values);
                       (e.g. segments_processed: 10000)
    """
    if not self._progress or self._part == 1:
      return
    data = self._data.get(symbol, None)
    if data is None:
      return
    data._counter += progress
    if data._counter == data._total:
      self.progress_end(symbol)
    elif data._partsize == 0 or \
         int(data._counter / data._partsize) > data._lastpart:
      if data._partsize == 0 and self._part > 0:
        return
      # this means total is very small
      if data._partsize > 0:
        data._lastpart = data._counter / data._partsize
      done = data._counter / data._total
      t = int(time.time() - data._starttime)
      eta = int((t / done) - t)
      tstr= ("Elapsed: {:02d}h {:02d}min {:02d}s"
             .format(t//3600, t//60%60, t%60))
      etastr = ("ETA: {:02d}h {:02d}min {:02d}s"
                .format(eta//3600, eta//60%60, eta%60))
      donestr = "{:.1f}".format(done*100)
      keystr = "".join([ "; {}: {}".format(k, v) for k,v in keyargs.items()])
      string = "{} {}% {} processed " \
               .format(self._pfx, donestr, data._units) + \
               "[{}; {}{}]" \
               .format(tstr, etastr, keystr)
      if len(string) > data._strlen:
        data._strlen = len(string)
        spacediff = ""
      else:
        spacediff = " "*(data._strlen-len(string))
      self._channel.write("\r{}{}".format(string, spacediff))
      self._channel.flush()

  def progress_end(self, symbol, **keyargs):
    """Completes progress logging for a computation.

    A message is always output. The progress is set to 100%.

    Parameters:
      symbol (str) : the identifier assigned to the computation when
                    `progress_init` was called
      **keyargs (dict) : additional units of computation to display (keys),
                       together with their final value (values);
                       (e.g. segments_processed: 100000)
    """
    if not self._progress:
      return
    data = self._data.get(symbol, None)
    if data is None:
      return
    t = int(time.time() - data._starttime)
    tstr= ("Elapsed time: {:02d}h {:02d}min {:02d}s"
           .format(t//3600, t//60%60, t%60))
    quantity = str(data._total) if self._part == 1 else "100.0%"
    keystr = "".join([ "; {}: {}".format(k, v) for k,v in keyargs.items()])
    string = "{} {} {} processed " \
             .format(self._pfx, quantity, data._units) + \
             "[{}{}]" \
             .format(tstr, keystr)
    spacediff = " " * (max(data._strlen - len(string),0))
    if self._part != 1:
      self._channel.write("\r")
    self._channel.write("{}{}\n".format(string, spacediff))
    self._channel.flush()
    self._data.pop(symbol)



import gfapy

class RGFA():
  """
  Add support of rGFA format.
  A dialect-specific validation method is added, as well as convenience
  methods to handle the stable sequence names.
  """

  def is_rgfa(self):
    """
    Indicate that rGFA dialect of GFA1 shall be used
    """
    return self._dialect == "rgfa"

  @property
  def stable_sequence_names(self):
    """Stable sequence names from rGFA SN tags"""
    if self._dialect != "rgfa":
      return []
    stable_seqs = set()
    for s in self.segments:
      stable_seqs.add(s.SN)
    return list(stable_seqs)

  def validate_rgfa(self):
    """
    Validate rGFA

    - version must be 1.0
    - no H, P, C lines are present
    - S lines have rGFA-specific predefined tags
    - if L lines have rGFA-specific tags, they have the correct type
    - overlaps must be 0M
    """
    self._validate_rgfa_version()
    self._validate_rgfa_no_headers()
    self._validate_rgfa_no_containments()
    self._validate_rgfa_no_paths()
    self._validate_rgfa_tags_in_lines(self.segments)
    self._validate_rgfa_tags_in_lines(self.dovetails)
    self._validate_rgfa_link_overlaps()

  def _validate_rgfa_version(self):
    """Validate version of rGFA (it must be gfa1)"""
    if self.version != "gfa1":
      raise gfapy.VersionError("rGFA format only supports GFA version 1")

  def _validate_rgfa_no_headers(self):
    """Validate the absence of H lines in rGFA"""
    if self.headers:
      raise gfapy.ValueError("rGFA does not support header lines")

  def _validate_rgfa_no_containments(self):
    """Validate the absence of C lines in rGFA"""
    if self.containments:
      raise gfapy.ValueError("rGFA does not support containment lines")

  def _validate_rgfa_no_paths(self):
    """Validate the absence of P lines in rGFA"""
    if self.paths:
      raise gfapy.ValueError("rGFA does not support path lines")

  RGFA_TAGS = {
    "mandatory": {
        "S": {"SN": "Z", "SO": "i", "SR": "i"},
        "L": {},
      },
    "optional": {
      "S": {},
      "L": {"SR": "i", "L1": "i", "L2": "i"},
      },
    }

  def _validate_rgfa_tags_in_lines(self, lines):
    """
    Validate rGFA tags for a group of lines
    """
    for line in lines:
      rt = line.record_type
      tags_check_presence = gfapy.Gfa.RGFA_TAGS["mandatory"].get(rt, {})
      tags_check_datatype = tags_check_presence.copy()
      tags_check_datatype.update(gfapy.Gfa.RGFA_TAGS["optional"].get(rt,{}))
      for tag, datatype in tags_check_presence.items():
        if tag not in line.tagnames:
          raise gfapy.NotFoundError(
            "rGFA {} lines must have a {} tag\n".format(rt, tag)+
            "offending line:\n{}".format(str(line)))
      for tag, datatype in tags_check_datatype.items():
        if tag in line.tagnames:
          if line.get_datatype(tag) != datatype:
            raise gfapy.ValueError(
              "rGFA {} tags in {} lines must have datatype {}\n".format(
                tag, rt, datatype)+
              "offending line:\n{}".format(str(line)))

  def _validate_rgfa_link_overlaps(self):
    for link in self.dovetails:
      if link.field_to_s("overlap") != "0M":
        raise gfapy.ValueError("rGFA CIGARs must be 0M\n",
              "offending line:\n{}".format(str(link)))



import re
import gfapy
from .alignment import Alignment

class CIGAR(list):
  """
  Representation of the contents of a CIGAR string.

  Each operation is represented by a
  :class:`CIGAR.Operation <gfapy.alignment.cigar.CIGAR.Operation>`,
  which specifies an operation length and operation symbol.

  Instances are usually created from their string representations, using the
  :class:`~gfapy.alignment.alignment.Alignment` factory class constructor.

  Warning:
    Although the GFA1 specification does not forbid the
    operation symbols NSHX=, these are not allowed in GFA2
    and thus their use in GFA1 is discouraged.
  """

  def complement(self):
    """The CIGAR when switching the role of the two aligned segments.

    Example:
      >>> import gfapy
      >>> str(gfapy.Alignment("2M1D3M").complement())
      '3M1I2M'

    Returns:
      CIGAR: the complement CIGAR
    """
    comp = list(reversed(self))
    for op in comp:
      if   op.code == "I": op.code = "D"
      elif op.code == "S": op.code = "D"
      elif op.code == "D": op.code = "I"
      elif op.code == "N": op.code = "I"
    return CIGAR(comp)

  def validate(self, version = "gfa1"):
    """Validates the instance.

    Parameters:
      version (str): 'gfa1' or 'gfa2'

    Raises:
      ~gfapy.error.VersionError: If a wrong **version** is specified.
      ~gfapy.error.TypeError: If a component of the list is not a
          CIGAR Operation; If the CIGAR operation length is not an integer or
          a string representing an integer.
      ~gfapy.error.ValueError: If the length of an operation is < 0; If an
          operation code is invalid in general or for the specified GFA version.
    """
    if version != "gfa1" and version != "gfa2":
      raise gfapy.VersionError(
          "Version error: {}".format(repr(version)))
    for op in self:
      if not isinstance(op, gfapy.CIGAR.Operation):
        raise gfapy.TypeError(
            "Element is not a CIGAR operation: {}\n".format(op)+
            "CIGAR instance is invalid: {}".format(self))
      op.validate(version = version)

  def length_on_reference(self):
    """Length of the aligned substring on the reference sequence
    (**from** sequence for GFA1 links/containments;
    **sid1** sequence for GFA2 edges)

    Returns:
      int
    """
    l = 0
    for op in self:
      if op.code in ["M", "=", "X", "D" , "N"]:
        l += op.length
    return l

  def length_on_query(self):
    """
    Lenght of the aligned substring on the query sequence
    (**to** sequence for GFA1 links/containments;
    **sid2** sequence for GFA2 edges)

    Returns:
      int
    """
    l = 0
    for op in self:
      if op.code in ["M", "=", "X", "I", "S"]:
        l += op.length
    return l

  @classmethod
  def _from_string(cls, string, valid = False, version = "gfa1"):
    """Create a CIGAR instance from its string representation.

    Parameters:
      string (str)
      valid (bool): If **True** the string is guaranteed to be valid.
        (Defaults to **False**)
      version (str): 'gfa1' or 'gfa2'

    Returns:
      ~gfapy.alignment.cigar.CIGAR or
      ~gfapy.alignment.placeholder.AlignmentPlaceholder

    Raises:
      ~gfapy.error.FormatError: If the string is not a valid CIGAR string.
    """
    if string == "*":
      return gfapy.AlignmentPlaceholder()
    cigar = CIGAR()
    if not valid:
      if version == "gfa1":
        if not re.match(r"^([0-9]+[MIDNSHPX=])+$", string):
          raise gfapy.FormatError()
      elif version == "gfa2":
        if not re.match(r"^([0-9]+[MIDP])+$", string):
          raise gfapy.FormatError()
    for m in re.finditer("([0-9]+)([MIDNSHPX=])", string):
      cigar.append(CIGAR.Operation(int(m.group(1)), m.group(2)))
    return cigar

  def __str__(self):
    if not self:
      return "*"
    else:
      return "".join([str(op) for op in self])

  def __repr__(self):
    return "gfapy.CIGAR([{}])".format(", ".join([repr(op) for op in self]))

  class Operation:
    """An operation in a CIGAR string.

    Attributes:
      ~Operation.length (int): Operation length.
      code (str): Operation code, one of
          :attr:`~Operation.CODE`.
    """

    CODE_GFA1_ONLY = ["S", "H", "N", "X", "="]
    """Operations only valid in GFA1"""

    CODE_GFA1_GFA2 = ["M", "I", "D", "P"]
    """Operations valid in GFA1 and GFA2"""

    CODE = CODE_GFA1_ONLY + CODE_GFA1_GFA2
    """CIGAR operation codes"""

    def validate(self, version = "gfa1"):
      """Validates the CIGAR operation.

      Parameters:
        version (str): 'gfa1' or 'gfa2'

      Raises:
        ~gfapy.error.VersionError: If a wrong **version** is specified.
        ~gfapy.error.TypeError: If the CIGAR operation length is not an integer
            or a string representing an integer.
        ~gfapy.error.ValueError: If the length of an operation is < 0; If an
            operation code is invalid in general or for the specified GFA
            version.
      """
      if version != "gfa1" and version != "gfa2":
        raise gfapy.VersionError(
            "Version error: {}".format(repr(version)))
      if not isinstance(self.length, int) and not isinstance(self.length, str):
        raise gfapy.TypeError(
            "Type error: length of CIGAR is {}".format(self.length))
      if(int(self.length) < 0):
        raise gfapy.ValueError("Length of CIGAR is {}".format(self.length))
      if version == "gfa2":
        if not self.code in Operation.CODE_GFA1_GFA2:
          raise gfapy.ValueError()
      else:
        if not self.code in Operation.CODE:
          raise gfapy.ValueError()

    def __init__(self, length, code):
      self.length = length
      self.code = code

    def __len__(self):
      return self.length

    def __str__(self):
      return "{}{}".format(self.length, self.code)

    def __repr__(self):
      return "gfapy.CIGAR.Operation({},{})".format(self.length, repr(self.code))

    def __eq__(self, other):
      return self.length == other.length and self.code == other.code

Operation = CIGAR.Operation



import gfapy
import unittest

class TestApiVersion(unittest.TestCase):

  def test_init_without_version_by_init(self):
    gfa = gfapy.Gfa()
    self.assertEqual(None, gfa.version)

  def test_init_GFA1(self):
    gfa = gfapy.Gfa(version="gfa1")
    self.assertEqual("gfa1", gfa.version)

  def test_init_GFA2(self):
    gfa = gfapy.Gfa(version="gfa2")
    self.assertEqual("gfa2", gfa.version)

  def test_init_invalid_version(self):
    self.assertRaises(gfapy.VersionError, gfapy.Gfa, version="x.x")

  def test_GFA1_header(self):
    hother = "H\taa:A:a\tff:f:1.1"
    hv1 = "H\tzz:Z:test\tVN:Z:1.0\tii:i:11"
    gfa = gfapy.Gfa()
    gfa.add_line(hother)
    self.assertEqual(None, gfa.version)
    gfa.add_line(hv1)
    self.assertEqual("gfa1", gfa.version)


  def test_GFA2_header(self):
    hother = "H\taa:A:a\tff:f:1.1"
    hv2 = "H\tzz:Z:test\tVN:Z:2.0\tii:i:11"
    gfa = gfapy.Gfa()
    gfa.add_line(hother)
    self.assertEqual(None, gfa.version)
    gfa.add_line(hv2)
    self.assertEqual("gfa2", gfa.version)


  def test_unknown_version_in_header(self):
    hother = "H\taa:A:a\tff:f:1.1"
    hvx = "H\tzz:Z:test\tVN:Z:x.x\tii:i:11"
    gfa = gfapy.Gfa()
    gfa.add_line(hother)
    self.assertEqual(None, gfa.version)
    self.assertRaises(gfapy.VersionError, gfa.add_line, hvx)


  def test_wrong_version_in_header(self):
    hother = "H\taa:A:a\tff:f:1.1"
    hv2 = "H\tzz:Z:test\tVN:Z:2.0\tii:i:11"
    gfa = gfapy.Gfa(version="gfa1")
    gfa.add_line(hother)
    self.assertEqual("gfa1", gfa.version)
    self.assertRaises(gfapy.VersionError, gfa.add_line, hv2)

  def test_conflicting_versions_in_header(self):
    hother = "H\taa:A:a\tff:f:1.1"
    hv1 = "H\tzz:Z:test\tVN:Z:1.0\tii:i:11"
    hv2 = "H\tzz:Z:test\tVN:Z:2.0\tii:i:11"
    gfa = gfapy.Gfa()
    gfa.add_line(hother)
    gfa.add_line(hv1)
    self.assertRaises(gfapy.VersionError, gfa.add_line, hv2)

  def test_version_by_segment_GFA1_syntax(self):
    sv1 = "S\tA\t*"
    gfa = gfapy.Gfa()
    gfa.add_line(sv1)
    self.assertEqual("gfa1", gfa.version)

  def test_version_by_segment_GFA2_syntax(self):
    sv2 = "S\tB\t100\t*"
    gfa = gfapy.Gfa()
    gfa.add_line(sv2)
    self.assertEqual("gfa2", gfa.version)

  def test_GFA2_segment_in_GFA1(self):
    sv1 = "S\tA\t*"
    sv2 = "S\tB\t100\t*"
    gfa = gfapy.Gfa()
    gfa.add_line(sv1)
    self.assertRaises(gfapy.VersionError, gfa.add_line, sv2)

  def test_GFA1_segment_in_GFA2(self):
    sv1 = "S\tA\t*"
    sv2 = "S\tB\t100\t*"
    gfa = gfapy.Gfa()
    gfa.add_line(sv2)
    self.assertRaises(gfapy.VersionError, gfa.add_line, sv1)

  def test_version_by_GFA2_specific_line_E(self):
    e = "E\t*\tA+\tB+\t0\t10\t20\t30\t*"
    gfa = gfapy.Gfa()
    gfa.add_line(e)
    self.assertEqual("gfa2", gfa.version)

  def test_version_by_GFA2_specific_line_G(self):
    g = "G\t*\tA+\tB-\t1000\t*"
    gfa = gfapy.Gfa()
    gfa.add_line(g)
    self.assertEqual("gfa2", gfa.version)

  def test_version_by_GFA2_specific_line_F(self):
    f = "F\tX\tID+\t10\t100\t0\t90$\t*"
    gfa = gfapy.Gfa()
    gfa.add_line(f)
    self.assertEqual("gfa2", gfa.version)

  def test_version_by_GFA2_specific_line_O(self):
    o = "O\tX\tA+ B- C+"
    gfa = gfapy.Gfa()
    gfa.add_line(o)
    self.assertEqual("gfa2", gfa.version)

  def test_version_by_GFA2_specific_line_U(self):
    u = "U\tX\tA B C"
    gfa = gfapy.Gfa()
    gfa.add_line(u)
    self.assertEqual("gfa2", gfa.version)

  def test_version_guess_GFA1_specific_line_L(self):
    string = "L\tA\t-\tB\t+\t*"
    gfa = gfapy.Gfa()
    gfa.add_line(string)
    gfa.process_line_queue()
    self.assertEqual("gfa1", gfa.version)

  def test_version_guess_GFA1_specific_line_C(self):
    string = "C\tA\t+\tB\t-\t10\t*"
    gfa = gfapy.Gfa()
    gfa.add_line(string)
    gfa.process_line_queue()
    self.assertEqual("gfa1", gfa.version)

  def test_version_guess_GFA1_specific_line_P(self):
    string = "P\t1\ta-,b+\t*"
    gfa = gfapy.Gfa()
    gfa.add_line(string)
    gfa.process_line_queue()
    self.assertEqual("gfa1", gfa.version)

  def test_version_guess_default(self):
    gfa = gfapy.Gfa()
    gfa.process_line_queue()
    self.assertEqual("gfa2", gfa.version)

  def test_header_version(self):
    self.assertEqual("generic", gfapy.Line("H\tVN:Z:1.0").version)
    self.assertEqual("gfa1", gfapy.Line("H\tVN:Z:1.0", version="gfa1").version)
    self.assertEqual("gfa2", gfapy.Line("H\tVN:Z:1.0", version="gfa2").version)

  def test_comment_version(self):
    self.assertEqual("generic", gfapy.Line("# VN:Z:1.0").version)
    self.assertEqual("gfa1", gfapy.Line("# VN:Z:1.0", version="gfa1").version)
    self.assertEqual("gfa2", gfapy.Line("# VN:Z:1.0", version="gfa2").version)

  def test_segment_version(self):
    self.assertEqual("gfa1", gfapy.Line("S\tA\tNNNN").version)
    self.assertEqual("gfa2", gfapy.Line("S\tA\t1\tNNNN").version)
    self.assertEqual("gfa1", gfapy.Line("S\tA\tNNNN", version="gfa1").version)
    self.assertEqual("gfa2", gfapy.Line("S\tA\t1\tNNNN", version="gfa2").version)
    self.assertRaises(gfapy.FormatError,
      gfapy.Line, "S\tA\t1\tNNNN", version="gfa1")
    self.assertRaises(gfapy.FormatError,
      gfapy.Line, "S\tA\tNNNN", version="gfa2")

  def test_link_version(self):
    string = "L\tA\t+\tB\t-\t*"
    self.assertEqual("gfa1", gfapy.Line(string).version)
    self.assertEqual("gfa1", gfapy.Line(string, version="gfa1").version)
    self.assertRaises(gfapy.VersionError, gfapy.Line, string, version="gfa2")
    self.assertRaises(gfapy.VersionError,
      gfapy.line.edge.Link, ["A","+","B","-","*"], version="gfa2")

  def test_containment_version(self):
    string = "C\tA\t+\tB\t-\t10\t*"
    self.assertEqual("gfa1", gfapy.Line(string).version)
    self.assertEqual("gfa1", gfapy.Line(string,version="gfa1").version)
    self.assertRaises(gfapy.VersionError, gfapy.Line,string,version="gfa2")
    self.assertRaises(gfapy.VersionError, gfapy.line.edge.Containment,
                      ["A","+","B","-","10","*"], version="gfa2")

  def test_edge_version(self):
    self.assertEqual("gfa2", gfapy.Line("E\t*\tA-\tB+\t0\t100\t0\t100\t*").version)
    self.assertEqual("gfa2", gfapy.Line("E\t*\tA-\tB+\t0\t100\t0\t100\t*",version=\
                                                            "gfa2").version)
    self.assertRaises(gfapy.VersionError,
      gfapy.Line, "E\t*\tA-\tB+\t0\t100\t0\t100\t*", version="gfa1")
    self.assertRaises(gfapy.VersionError,
      gfapy.line.edge.GFA2, ["A-","B+", "0", "100", "0", "100", "*"],
                                 version="gfa1")

  def test_gap_version(self):
    self.assertEqual("gfa2", gfapy.Line("G\t*\tA-\tB+\t100\t*").version)
    self.assertEqual("gfa2", gfapy.Line("G\t*\tA-\tB+\t100\t*",
                               version="gfa2").version)
    self.assertRaises(gfapy.VersionError,
      gfapy.Line, "G\t*\tA-\tB+\t100\t*", version="gfa1")
    self.assertRaises(gfapy.VersionError,
      gfapy.line.Gap,["A-","B+", "100", "*"], version="gfa1")

  def test_fragment_version(self):
    self.assertEqual("gfa2", gfapy.Line("F\tA\tread1-\t0\t100\t0\t100\t*").version)
    self.assertEqual("gfa2", gfapy.Line("F\tA\tread1-\t0\t100\t0\t100\t*", version=\
                                                            "gfa2").version)
    self.assertRaises(gfapy.VersionError,
      gfapy.Line, "F\tA\tread1-\t0\t100\t0\t100\t*", version="gfa1")
    self.assertRaises(gfapy.VersionError,
      gfapy.line.Fragment,["A","read-", "0", "100", "0", "100", "*"],
                               version="gfa1")

  def test_custom_record_version(self):
    self.assertEqual("gfa2", gfapy.Line("X\tVN:Z:1.0").version)
    self.assertEqual("gfa2", gfapy.Line("X\tVN:Z:1.0", version="gfa2").version)
    self.assertRaises(gfapy.VersionError,
      gfapy.Line, "X\tVN:Z:1.0", version="gfa1")
    self.assertRaises(gfapy.VersionError,
      gfapy.line.CustomRecord, ["X","VN:Z:1.0"], version="gfa1")

  def test_path_version(self):
    string = "P\t1\tA+,B-\t*"
    self.assertEqual("gfa1", gfapy.Line(string).version)
    self.assertEqual("gfa1", gfapy.Line(string, version="gfa1").version)
    self.assertRaises(gfapy.VersionError, gfapy.Line, string, version="gfa2")
    string = "O\t1\tA+ B-"
    self.assertEqual("gfa2", gfapy.Line(string).version)
    self.assertEqual("gfa2", gfapy.Line(string, version="gfa2").version)
    self.assertRaises(gfapy.VersionError, gfapy.Line, string, version="gfa1")

  def test_set_version(self):
    string = "U\t1\tA B C"
    self.assertEqual("gfa2", gfapy.Line(string).version)
    self.assertEqual("gfa2", gfapy.Line(string, version="gfa2").version)
    self.assertRaises(gfapy.VersionError, gfapy.Line, string, version="gfa1")

  def test_unknown_record_version(self):
    self.assertEqual("gfa2", gfapy.line.Unknown([None, "A"]).version)
    self.assertEqual("gfa2", gfapy.line.Unknown([None, "A"], version="gfa2").version)
    self.assertRaises(gfapy.VersionError, gfapy.line.Unknown,["\n","A"], version="gfa1")



import gfapy
import unittest

class TestLineLink(unittest.TestCase):

  def test_from_string(self):
    fields=["L","1","+","2","-","12M","RC:i:1232","NM:i:3","ab:Z:abcd",
            "FC:i:2321","KC:i:1212","MQ:i:40"]
    string = "\t".join(fields)
    gfapy.Line(string)
    self.assertIsInstance(gfapy.Line(string), gfapy.line.edge.Link)
    self.assertEqual(str(fields[0]), gfapy.Line(string).record_type)
    self.assertEqual(str(fields[1]), gfapy.Line(string).from_segment)
    self.assertEqual(str(fields[2]), gfapy.Line(string).from_orient)
    self.assertEqual(str(fields[3]), gfapy.Line(string).to_segment)
    self.assertEqual(str(fields[4]), gfapy.Line(string).to_orient)
    self.assertEqual([gfapy.alignment.CIGAR.Operation(12, "M")],
                      gfapy.Line(string).overlap)
    self.assertEqual(1232, gfapy.Line(string).RC)
    self.assertEqual(3, gfapy.Line(string).NM)
    self.assertEqual(2321, gfapy.Line(string).FC)
    self.assertEqual(1212, gfapy.Line(string).KC)
    self.assertEqual(40, gfapy.Line(string).MQ)
    self.assertEqual("abcd", gfapy.Line(string).ab)
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line((string+"\tH1"))
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line("L\tH")
    with self.assertRaises(gfapy.FormatError):
      f=fields[:]
      f[2]="x"
      gfapy.Line("\t".join(f), vlevel = 2)
    with self.assertRaises(gfapy.FormatError):
      f=fields[:]
      f[4]="x"
      gfapy.Line("\t".join(f), vlevel = 2)
    with self.assertRaises(gfapy.FormatError):
      f=fields[:]
      f[5]="x"
      gfapy.Line("\t".join(f), vlevel = 2)
    with self.assertRaises(gfapy.TypeError):
      f=fields[:]
      f[6]="RC:Z:1232"
      gfapy.Line("\t".join(f), vlevel = 2)
    with self.assertRaises(gfapy.TypeError):
      f=fields[:]
      f[7]="NM:Z:1232"
      gfapy.Line("\t".join(f), vlevel = 2)

  #TODO
  #def test_coords
  #  g = RGFA.new(version: :gfa1)
  #  g << "S\t1\t*\tLN:i:100"
  #  g << "L\t1\t+\t2\t-\t1M2D10M1I"
  #  assert_equal([87,100], g.links[0].from_coords)
  #  assert_raises(RGFA::ValueError) {g.links[0].to_coords}
  #  g << "S\t2\t*\tLN:i:100"
  #  assert_equal([88,100], g.links[0].to_coords)
  #  g << "L\t3\t-\t4\t+\t10M2P3D1M"
  #  assert_equal([0,14], g.links[1].from_coords)
  #  assert_equal([0,11], g.links[1].to_coords)
  #end

  #def test_to_gfa2
  #  g = RGFA.new(version: :gfa1)
  #  g << "S\t1\t*\tLN:i:100"
  #  g << "S\t2\t*\tLN:i:100"
  #  g << "S\t3\t*\tLN:i:100"
  #  g << "S\t4\t*\tLN:i:100"
  #  g << "L\t1\t+\t2\t+\t10M"
  #  g << "L\t1\t-\t2\t-\t20M"
  #  g << "L\t3\t-\t4\t+\t30M"
  #  g << "L\t3\t+\t4\t-\t40M"
  #  assert_equal("E	*	1+	2+	90	100$	0	10	10M",
  #               g.links[0].to_gfa2_s)
  #  assert_equal("E	*	1-	2-	0	20	80	100$	20M",
  #               g.links[1].to_gfa2_s)
  #  assert_equal("E	*	3-	4+	0	30	0	30	30M",
  #               g.links[2].to_gfa2_s)
  #  assert_equal("E	*	3+	4-	60	100$	60	100$	40M",
  #               g.links[3].to_gfa2_s)
  #  assert_equal(RGFA::Line::Edge::Link, g.links[0].to_gfa1.class)
  #  assert_equal(RGFA::Line::Edge::GFA2, g.links[0].to_gfa2.class)
  #end



import gfapy
try:
  from functools import partialmethod
except ImportError:
  #for compatibility with old python versions
  def partialmethod(method, **kwargs):
    return lambda self: method(self, **kwargs)

class VersionConversion:

  @property
  def version(self):
    """
    Returns
    -------
    gfapy.VERSIONS, None
      GFA specification version
    """
    return self._version

  @property
  def dialect(self):
    """
    Returns
    -------
    gfapy.DIALECTS, None
      GFA specification version
    """
    return self._dialect

  def to_version_s(self, version):
    """
    Returns
    -------
    str
      A string representation of self.
    """
    return gfapy.Line.SEPARATOR.join(getattr(self, "_to_"+version+"_a")())

  def _to_version_a(self, version):
    """
    .. note::
      The default is an alias of to_list() if version is equal to the
      version of the line, and an empty list otherwise.
      gfapy.Line subclasses can redefine this method to convert
      between versions.

    Returns
    -------
    str list
      A list of string representations of the fields.
    """
    if version == self._version:
      return self.to_list()
    else:
      return []

  def to_version(self, version, raise_on_failure=True):
    """
    Returns
    -------
    gfapy.Line
    	Conversion to the selected version.
    """
    if version == self._version:
      return self
    elif version not in gfapy.VERSIONS:
      raise gfapy.VersionError("Version unknown ({})".format(version))
    else:
      l = getattr(self, "_to_"+version+"_a")()
      if l:
        try:
          converted = gfapy.Line(l, version=version, vlevel=self.vlevel)
        except:
          raise gfapy.RuntimeError("Conversion to {} failed\n".format(version)+
              "Line: {}".format(str(self)))
        return converted
      elif raise_on_failure:
        raise gfapy.VersionError("Records of type {} ".format(self.record_type)+
            "cannot be converted from version {} ".format(self._version)+
            "to version {}".format(version))
      else:
        return None

for shall_version in ["gfa1", "gfa2"]:
  setattr(VersionConversion, "to_"+shall_version+"_s",
          partialmethod(VersionConversion.to_version_s,
            version = shall_version))

  setattr(VersionConversion, "_to_"+shall_version+"_a",
          partialmethod(VersionConversion._to_version_a,
            version = shall_version))

  setattr(VersionConversion, "to_"+shall_version,
          partialmethod(VersionConversion.to_version,
            version = shall_version))



import gfapy

class SuperfluousLinks:

  def enforce_segment_mandatory_links(self, segment, conserve_components=True):
    """Enforce mandatory dovetails overlaps of a given segment to other
    segments, by removing all other dovetail overlaps between those segments.

    The definition of mandatory links follows the one
    given in Gonnella and Kurtz, 2016.

    Parameters:
      segment (Line, str) : the segment
      conserve_components (bool): if True, then dovetail overlaps are only
        removed, if their removal does not split connected components
        of the graph (considering as connections only dovetail overlaps)
    """
    s, sn = self._segment_and_segment_name(segment)
    se = {}; l = {}
    for et in ["L", "R"]:
      se[et] = [sn, et]
      l[et] = segment(s).dovetails_of_end(et)
    if len(l["L"])==1 and len(l["R"])==1:
      oe = {}
      for et in ["L", "R"]:
        oe[et] = l[et][0].other_end(se[et])
      if oe[:L] == oe[:R]:
        return
      for et in ["L", "R"]:
        self._delete_other_links(oe[et], se[et],
                                conserve_components=conserve_components)
    else:
      if l[:L].size == 1:
        et = "L"
      elif l[:R].size == 1:
        et = "R"
      else:
        return
      oe = l[et][0].other_end(se[et])
      self._delete_other_links(oe, se[et],
                              conserve_components=conserve_components)

  def enforce_all_mandatory_links(self, conserve_components=True):
    """Enforce mandatory dovetails between pairs of segments, by removing all
       other dovetail overlaps between those segments.

    The definition of mandatory links follows the one
    given in Gonnella and Kurtz, 2016.

    Parameters:
      conserve_components (bool): if True, then dovetail overlaps are only
        removed, if their removal does not split connected components
        of the graph (considering as connections only dovetail overlaps)
    """
    for sn in segment_names:
      self.enforce_segment_mandatory_links(sn, conserve_components=
                                               conserve_components)

  def remove_self_link(self, segment):
    """Remove self links of a segment, if any.

    Remove any dovetail overlap of a segment to itself.

    Parameters:
      segment (str, Line): the segment
    """
    if not isinstance(segment, gfa.Line):
      segment = self.try_get_segment(segment)
    for e in segment.dovetails:
      if e.from_segment == e.to_segment:
        e.disconnect()

  def remove_self_links(self):
    """Remove all dovetail overlap of segments to themselves, if any."""
    for sn in segment_names:
      self.remove_self_link(sn)



import unittest
import gfapy

class TestUnitLineEquivalence(unittest.TestCase):

  a      = gfapy.Line("S\tA\t*\tLN:i:8\txx:Z:a")
  b      = gfapy.Line("S\tB\t*\tLN:i:10")
  c      = gfapy.Line("C\tA\t+\tB\t+\t10\t*")
  l      = gfapy.Line("L\tA\t+\tB\t+\t*")
  e      = gfapy.Line("E\t1\tA+\tB-\t0\t100$\t20\t121\t*")

  a_ln   = gfapy.Line("S\tA\t*\tLN:i:10\txx:Z:a")
  a_seq  = gfapy.Line("S\tA\tACCTTCGT\tLN:i:8\txx:Z:a")
  a_gfa2 = gfapy.Line("S\tA\t8\tACCTTCGT\txx:Z:a")
  a_noxx = gfapy.Line("S\tA\t*\tLN:i:8")
  a_yy   = gfapy.Line("S\tA\t*\tLN:i:8\txx:Z:a\tyy:Z:b")
  l_from = gfapy.Line("L\tC\t+\tB\t+\t*")
  e_name = gfapy.Line("E\t2\tA+\tB-\t0\t100$\t20\t121\t*")

  h_a    = {"record_type": "S",
            "name": "A",
            "LN": 8,
            "xx": "a"}
  h_a_rt = h_a.copy()
  h_a_rt["record_type"] = "X"
  h_a_pl = h_a.copy()
  h_a_pl["name"] = gfapy.Placeholder()
  h_a_name = h_a.copy()
  h_a_name["name"] = "B"
  h_a_seq = h_a.copy()
  h_a_seq["sequence"] = "ACCTTCGT"
  h_a_ln = h_a.copy()
  h_a_ln["LN"] = 10
  h_a_LNstr = h_a.copy()
  h_a_LNstr["LN"] = "8"
  h_a_noxx = h_a.copy()
  h_a_noxx.pop("xx")
  h_a_yy = h_a.copy()
  h_a_yy["yy"] = "b"
  h_a_gfa2 = {"record_type": "S",
              "sid": "A",
              "slen": 8,
              "xx": "a"}


  def test_line_placeholder(self):
    assert(not gfapy.is_placeholder(TestUnitLineEquivalence.a))
    assert(not gfapy.is_placeholder(TestUnitLineEquivalence.b))

  def test_line_diff_two_segments(self):
    adiffb = [("different", "positional_field", "name", "A", "B"),
              ("exclusive", "<", "tag", "xx", "Z", "a"),
              ("different", "tag", "LN", "i", "8", "i", "10")]
    self.assertEqual(sorted(adiffb), sorted(TestUnitLineEquivalence.a.diff(TestUnitLineEquivalence.b)))
    bdiffa = [("different", "positional_field", "name", "B", "A"),
              ("exclusive", ">", "tag", "xx", "Z", "a"),
              ("different", "tag", "LN", "i", "10", "i", "8")]
    self.assertEqual(sorted(bdiffa), sorted(TestUnitLineEquivalence.b.diff(TestUnitLineEquivalence.a)))
    self.assertEqual([], TestUnitLineEquivalence.a.diff(TestUnitLineEquivalence.a))
    self.assertEqual([], TestUnitLineEquivalence.b.diff(TestUnitLineEquivalence.b))

  def test_line_diffscript_two_segments(self):
    acpy = TestUnitLineEquivalence.a.clone()
    exec(acpy.diffscript(TestUnitLineEquivalence.b, "acpy"))
    self.assertNotEqual(str(TestUnitLineEquivalence.b), str(TestUnitLineEquivalence.a))
    self.assertEqual(str(TestUnitLineEquivalence.b), str(acpy))
    bcpy = TestUnitLineEquivalence.b.clone()
    exec(bcpy.diffscript(TestUnitLineEquivalence.a, "bcpy"))
    self.assertNotEqual(str(TestUnitLineEquivalence.a), str(TestUnitLineEquivalence.b))
    self.assertEqual(str(TestUnitLineEquivalence.a), str(bcpy))

  def test_equal(self):
    assert(TestUnitLineEquivalence.a == TestUnitLineEquivalence.a)
    assert(TestUnitLineEquivalence.b == TestUnitLineEquivalence.b)
    assert(TestUnitLineEquivalence.c == TestUnitLineEquivalence.c)
    assert(TestUnitLineEquivalence.l == TestUnitLineEquivalence.l)
    assert(TestUnitLineEquivalence.e == TestUnitLineEquivalence.e)
    assert(not (TestUnitLineEquivalence.a == TestUnitLineEquivalence.b))
    assert(not (TestUnitLineEquivalence.a == TestUnitLineEquivalence.a_ln))
    assert(not (TestUnitLineEquivalence.a == TestUnitLineEquivalence.a_seq))
    assert(not (TestUnitLineEquivalence.a == TestUnitLineEquivalence.a_gfa2))
    assert(not (TestUnitLineEquivalence.a == TestUnitLineEquivalence.a_noxx))
    assert(TestUnitLineEquivalence.b == TestUnitLineEquivalence.b.clone())
    assert(TestUnitLineEquivalence.a == TestUnitLineEquivalence.a.clone())

  def test_pointer_equality(self):
    assert(TestUnitLineEquivalence.a is TestUnitLineEquivalence.a)
    assert(not TestUnitLineEquivalence.a is TestUnitLineEquivalence.a.clone())

  def test_has_eql_fields(self):
    # same object
    assert(TestUnitLineEquivalence.a._has_eql_fields(TestUnitLineEquivalence.a))
    # clone
    assert(TestUnitLineEquivalence.a._has_eql_fields(TestUnitLineEquivalence.a.clone()))
    # positional field difference
    assert(not TestUnitLineEquivalence.l._has_eql_fields(TestUnitLineEquivalence.l_from))
    assert(TestUnitLineEquivalence.l._has_eql_fields(TestUnitLineEquivalence.l_from, ["from"]))
    # positional field difference: name alias
    assert(not TestUnitLineEquivalence.e._has_eql_fields(TestUnitLineEquivalence.e_name))
    assert(TestUnitLineEquivalence.e._has_eql_fields(TestUnitLineEquivalence.e_name, ["eid"]))
    assert(TestUnitLineEquivalence.e._has_eql_fields(TestUnitLineEquivalence.e_name, ["name"]))
    # positional field difference: placeholder in line
    assert(TestUnitLineEquivalence.a._has_eql_fields(TestUnitLineEquivalence.a_seq))
    # positional field difference: placeholder in reference
    assert(TestUnitLineEquivalence.a_seq._has_eql_fields(TestUnitLineEquivalence.a))
    # tag difference
    assert(not TestUnitLineEquivalence.a._has_eql_fields(TestUnitLineEquivalence.a_ln))
    assert(TestUnitLineEquivalence.a._has_eql_fields(TestUnitLineEquivalence.a_ln, ["LN"]))
    # additional tag in line
    assert(TestUnitLineEquivalence.a._has_eql_fields(TestUnitLineEquivalence.a_noxx))
    assert(not TestUnitLineEquivalence.a_noxx._has_eql_fields(TestUnitLineEquivalence.a))
    # missing tag in line
    assert(not TestUnitLineEquivalence.a._has_eql_fields(TestUnitLineEquivalence.a_yy))
    assert(TestUnitLineEquivalence.a_yy._has_eql_fields(TestUnitLineEquivalence.a))
    assert(TestUnitLineEquivalence.a._has_eql_fields(TestUnitLineEquivalence.a_yy, ["yy"]))
    # gfa1 vs gfa2
    assert(TestUnitLineEquivalence.a._has_eql_fields(TestUnitLineEquivalence.a_gfa2, ["slen"]))
    assert(TestUnitLineEquivalence.a_gfa2._has_eql_fields(TestUnitLineEquivalence.a, ["LN"]))
    # record_type
    assert(not TestUnitLineEquivalence.c._has_eql_fields(TestUnitLineEquivalence.l))
    assert(not TestUnitLineEquivalence.l._has_eql_fields(TestUnitLineEquivalence.c))
    assert(TestUnitLineEquivalence.c._has_eql_fields(TestUnitLineEquivalence.l, ["record_type"]))
    assert(TestUnitLineEquivalence.l._has_eql_fields(TestUnitLineEquivalence.c, ["record_type", "pos"]))

  def test_has_field_values(self):
    assert(TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a))
    # record_type difference
    assert(not TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_rt))
    assert(TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_rt, ["record_type"]))
    # positional field difference
    assert(not TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_name))
    assert(TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_name, ["name"]))
    # positional field difference: placeholder in line
    assert(TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_seq))
    # positional field difference: placeholder in hash is compared
    assert(not TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_pl))
    assert(TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_pl, ["name"]))
    # tag difference
    assert(not TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_ln))
    assert(TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_ln, ["LN"]))
    # encoded value
    assert(TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_LNstr))
    # additional tag in line
    assert(TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_noxx))
    # missing tag in line
    assert(not TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_yy))
    assert(TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_yy, ["yy"]))
    # gfa1 vs gfa2
    assert(TestUnitLineEquivalence.a._has_field_values(TestUnitLineEquivalence.h_a_gfa2, ["slen"]))




import unittest
import gfapy

class TestUnitLineConnection(unittest.TestCase):

  def test_connected_and_gfa(self):
    s1 = gfapy.Line("S\t1\tACCAT")
    assert(not s1.is_connected())
    self.assertEqual(None, s1.gfa)
    g = gfapy.Gfa()
    g.append(s1)
    assert(s1.is_connected())
    assert(g is s1.gfa)

  def test_connect(self):
    s2 = gfapy.Line("S\t2\tACCAT")
    assert(not s2.is_connected())
    self.assertEqual(None, s2.gfa)
    g = gfapy.Gfa()
    s2.connect(g)
    assert(s2.is_connected())
    assert(g is s2.gfa)

  def test_connect_registers_line(self):
    s2 = gfapy.Line("S\t2\tACCAT")
    g = gfapy.Gfa()
    self.assertEqual([], g.segments)
    s2.connect(g)
    self.assertEqual([s2], g.segments)

  def test_disconnect(self):
    s1 = gfapy.Line("S\t1\tACCAT")
    g = gfapy.Gfa()
    g.append(s1)
    assert(s1.is_connected())
    assert(g is s1.gfa)
    s1.disconnect()
    assert(not s1.is_connected())
    self.assertEqual(None, s1.gfa)

  def test_disconnect_unregisters_line(self):
    s1 = gfapy.Line("S\t1\tACCAT")
    g = gfapy.Gfa()
    g.append(s1)
    self.assertEqual([s1], g.segments)
    s1.disconnect()
    self.assertEqual([], g.segments)

  def test_disconnect_removes_field_backreferences(self):
    s1 = gfapy.Line("S\t1\tACCAT")
    l = gfapy.Line("L\t1\t+\t2\t-\t*")
    g = gfapy.Gfa()
    g.append(s1)
    g.append(l)
    self.assertEqual([l], s1.dovetails)
    l.disconnect()
    self.assertEqual([], s1.dovetails)

  def test_disconnect_removes_field_references(self):
    s1 = gfapy.Line("S\t1\tACCAT")
    l = gfapy.Line("L\t1\t+\t2\t-\t*")
    g = gfapy.Gfa()
    g.append(s1)
    g.append(l)
    assert(l.get("from") is s1)
    l.disconnect()
    assert(not l.get("from") is s1)
    self.assertEqual("1", l.get("from"))

  def test_disconnect_disconnects_depent_lines(self):
    s1 = gfapy.Line("S\t1\tACCAT")
    l = gfapy.Line("L\t1\t+\t2\t-\t*")
    g = gfapy.Gfa()
    g.append(s1)
    g.append(l)
    assert(l.is_connected())
    s1.disconnect()
    assert(not l.is_connected())

  def test_disconnect_removes_nonfield_backreferences(self):
    s1 = gfapy.Line("S\t1\tACCAT")
    s2 = gfapy.Line("S\t2\tCATGG")
    s3 = gfapy.Line("S\t3\tTGGAA")
    l12 = gfapy.Line("L\t1\t+\t2\t+\t*")
    l23 = gfapy.Line("L\t2\t+\t3\t+\t*")
    p4 = gfapy.Line("P\t4\t1+,2+,3+\t*")
    g = gfapy.Gfa()
    for line in [s1, s2, s3, l12, l23, p4]:
      g.append(line)
    self.assertEqual([p4], l12.paths)
    p4.disconnect()
    self.assertEqual([], l12.paths)

  def test_disconnect_removes_nonfield_references(self):
    s1 = gfapy.Line("S\t1\tACCAT")
    s2 = gfapy.Line("S\t2\tCATGG")
    s3 = gfapy.Line("S\t3\tTGGAA")
    l12 = gfapy.Line("L\t1\t+\t2\t+\t*")
    l23 = gfapy.Line("L\t2\t+\t3\t+\t*")
    p4 = gfapy.Line("P\t4\t1+,2+,3+\t*")
    g = gfapy.Gfa()
    for line in [s1, s2, s3, l12, l23, p4]:
      g.append(line)
    self.assertEqual([gfapy.OrientedLine(l12,"+"),gfapy.OrientedLine(l23,"+")], p4.links)
    p4.disconnect()
    self.assertEqual([], p4.links)

  def test_add_reference(self):
    s1 = gfapy.Line("S\t1\tACCAT")
    self.assertEqual([], s1.gaps_L)
    s1._add_reference("X", "gaps_L")
    self.assertEqual(["X"], s1.gaps_L)
    s1._add_reference("Y", "gaps_L")
    self.assertEqual(["X", "Y"], s1.gaps_L)
    s1._add_reference("Z", "gaps_L", append=False)
    self.assertEqual(["Z", "X", "Y"], s1.gaps_L)

  def test_delete_reference(self):
    s1 = gfapy.Line("S\t1\tACCAT")
    s1._add_reference("A", "gaps_L")
    s1._add_reference("B", "gaps_L")
    s1._add_reference("C", "gaps_L")
    s1._add_reference("D", "gaps_L")
    s1._add_reference("E", "gaps_L")
    self.assertEqual(["A", "B", "C", "D", "E"], s1.gaps_L)
    s1._delete_reference("C", "gaps_L")
    self.assertEqual(["A", "B", "D", "E"], s1.gaps_L)
    s1._delete_first_reference("gaps_L")
    self.assertEqual(["B", "D", "E"], s1.gaps_L)
    s1._delete_last_reference("gaps_L")
    self.assertEqual(["B", "D"], s1.gaps_L)

  def test_update_references(self):
    s1 = gfapy.Line("S\t1\tACCAT")
    gA = gfapy.line.Gap({})
    gnewA = gfapy.line.Gap({})
    gB = gfapy.line.Gap({})
    gC = gfapy.line.Gap({})
    gD = gfapy.line.Gap({})
    gE = gfapy.line.Gap({})
    gX = gfapy.line.Gap({})
    s1._add_reference(gA, "gaps_L")
    s1._add_reference(gB, "gaps_L")
    s1._add_reference(gC, "gaps_L")
    s1._add_reference(gD, "gaps_L")
    s1._add_reference(gE, "gaps_L")
    self.assertEqual([gA, gB, gC, gD, gE], s1.gaps_L)
    s1._update_references(gA, gnewA, "sid1")
    self.assertEqual([gnewA, gB, gC, gD, gE], s1.gaps_L)
    s1._update_references(gX, "newX", "sid1")
    self.assertEqual([gnewA, gB, gC, gD, gE], s1.gaps_L)
    s1._update_references(gB, None, "sid1")
    self.assertEqual([gnewA, gC, gD, gE], s1.gaps_L)




"""
Update of references caused by a virtual line becoming real.
"""
import gfapy

class UpdateReferences:

  def _update_references(self, oldref, newref, key_in_ref):
    """
    This is called on lines which were referenced by virtual lines,
    when a real line is found which substitutes the virtual line.

    .. note::
      SUBCLASSES which can be referenced by virtual lines
      may implement a specialized *_backreference_keys* method to
      support this mechanism (the default will work in all cases
      of the current specification, but is not optimized for record type)

    Parameters
    ----------
    oldref : gfapy.Line
    newref : gfapy.Line
    key_in_ref : str list
    """
    keys = self._backreference_keys(oldref, key_in_ref)
    assert(keys is not None)
    self.__update_field_references(oldref, newref,
                                  list(set(self.__class__.REFERENCE_FIELDS)
                                       .intersection(keys)))
    if hasattr(self, "_refs"):
      # note: keeping the two types of nonfield references separate helps
      #       in subclasses where only one must be redefined
      self.__update_dependent_line_references(oldref, newref,
                                             set(self.__class__.DEPENDENT_LINES)
                                             .intersection(self._refs.keys())
                                             .intersection(keys))
      self.__update_other_references(oldref, newref,
                                    list(set(self.__class__.OTHER_REFERENCES)
                                    .intersection(self._refs.keys())
                                    .intersection(keys)))

  def _backreference_keys(self, ref, key_in_ref):
    """
    Return a list of fields and/or @ref keys, which indicates
    where a reference "ref" _may_ be stored (in order to be able
    to locate it and update it).

    The default is: all reference fields, dependent line references
    and other references.

    .. note::
      SUBCLASSES may overwrite this method if they
      can be referenced by virtual lines, by providing more
      specific results, depending on the ref and key_in_ref;
      this can make the update faster.

    Returns
    -------
    str list
      Fieldnames and/or _refs keys.
    """
    return (self.__class__.REFERENCE_FIELDS +
            self.__class__.DEPENDENT_LINES +
            self.__class__.OTHER_REFERENCES)

  def __update_reference_in_field(self, field, oldref, newref):
    """
    .. note::
      This methods supports fields which contain references,
      oriented lines or lists of references or oriented lines;
      if SUBCLASSES contain fields which reference to line in a
      different fashion, the method must be updated or overwritten
      by the subclass
    """
    value = self.get(field)
    if isinstance(value, gfapy.Line):
      if value is oldref:
        self._set_existing_field(field, newref, set_reference = True)
    elif isinstance(value, gfapy.OrientedLine):
      if value.line is oldref:
        value.line = newref
    elif isinstance(value, list):
      self.__update_reference_in_list(value, oldref, newref)

  def __update_reference_in_list(self, lst, oldref, newref):
    found = False
    for idx, elem in enumerate(lst):
      if isinstance(elem, gfapy.Line):
        if elem is oldref:
          lst[idx] = newref
          found = True
      elif isinstance(elem, gfapy.OrientedLine):
        if elem.line is oldref:
          if hasattr(oldref, "is_complement") and \
                            oldref.is_complement(newref):
            elem.orient = gfapy.invert(elem.orient)
          elem.line = newref
          found = True
    if newref is None and found:
      lst[:] = [e for e in lst if e is not None]

  def __update_field_references(self, oldref, newref, possible_fieldnames):
    for fn in possible_fieldnames:
      self.__update_reference_in_field(fn, oldref,
          newref if newref else str(oldref))

  def __update_nonfield_references(self, oldref, newref, possible_keys):
    for key in possible_keys:
      if key in self._refs:
        self.__update_reference_in_list(self._refs[key], oldref, newref)

  def __update_dependent_line_references(self ,oldref, newref, possible_keys):
    self.__update_nonfield_references(oldref, newref, possible_keys)

  def __update_other_references(self, oldref, newref, possible_keys):
    """
    .. note:: SUBCLASSES may redefine this method
    """
    self.__update_nonfield_references(oldref, newref, possible_keys)



class VersionConversion:

  def _to_gfa2_a(self):
    """
    Return the string representation of the tags, changing the value
    of the VN tag to 2.0, if this is present

    Returns
    -------
    list of str
    	Array of strings representing the tags.
    """
    a = ["H"]
    if self.VN:
      a.append("VN:Z:2.0")
    for fn in self.tagnames:
      if fn != "VN":
        a.append(self.field_to_s(fn, tag = True))
    return a

  def _to_gfa1_a(self):
    """
    Return the string representation of the tags, changing the value
    of the VN tag to 1.0, if this is present

    Returns
    -------
    list of str
    	Array of strings representing the tags.
    """
    a = ["H"]
    if self.VN:
      a.append("VN:Z:1.0")
    for fn in self.tagnames:
      if fn != "VN":
        a.append(self.field_to_s(fn, tag = True))
    return a



from .construction import Construction
from ..line import Line

class CustomRecord(Construction, Line):
  """Custom record of a GFA2 file.

  According to the specification, any line that does not begin with a
  recognized code can be ignored. This allows users to have additional
  descriptor lines specific to their special processes.

  Parsing of custom lines is handled as follows:
  - divide content by tabs
  - from the back, fields are parsed as GFA tags (XX:Y:...); until an exception
    is thrown, they are all considered tags
  - from the first exception back to the first field, they are all considered
    positional fields with name field1, field2, etc
  """

  RECORD_TYPE = None
  POSFIELDS = ["record_type"]
  DATATYPE = {
    "record_type" : "custom_record_type"
  }

CustomRecord._apply_definitions()



import gfapy

class References():

  def _initialize_references(self):
    s = self._gfa.segment(self.get("sid"))
    if s is None:
      if self._gfa._segments_first_order:
        raise gfapy.NotFoundError()
      s = gfapy.line.segment.GFA2({"sid": self.get("sid"),
                                   "slen": 1,
                                   "sequence": "*"},
                                   version = "gfa2",
                                   virtual = True)
      s.connect(self._gfa)
    self._set_existing_field("sid", s, set_reference = True)
    s._add_reference(self, "fragments")



from ..line import Line

class Unknown(Line):
  """
  A GFA2 line which was referred to only by G or O lines
  and has not been found yet (ie is always virtual)
  """

  RECORD_TYPE = "\n"
  POSFIELDS = ["name"]
  DATATYPE = {"name": "identifier_gfa2"}
  NAME_FIELD = "name"
  DEPENDENT_LINES = ["sets", "paths"]

  def __str__(self):
    return "?record_type?\t{}\tco:Z:line_created_by_gfapy".format(self.name)

  @property
  def virtual(self):
    return True

Unknown._apply_definitions()



import gfapy

class Creators:

  def add_line(self, gfa_line):
    """Add a line to a GFA instance.

    Note:
      append() is an alias to this method

    Parameters:
      gfa_line (str, Line): a line instance or a string, containing a line
        of a GFA file (if a string, a line instance is constructed using
        the string)

    Raises:
      gfapy.error.VersionError : If a wrong line type is used, for the GFA
        version
      gfapy.error.FormatError : If the content of the line string is
        not valid
    """
    if gfa_line is None:
      return
    if self._version == "gfa1":
      self.__add_line_GFA1(gfa_line)
    elif self._version == "gfa2":
      self.__add_line_GFA2(gfa_line)
    elif self._version is None:
      self.__add_line_unknown_version(gfa_line)
    else:
      raise gfapy.AssertionError("This point should never be reached")

  append = add_line

  def process_line_queue(self):
    """Process the lines kept by side while parsing GFA of unknown version.

    This method is called after adding the lines, if the GFA version is
    not specified, at soon as the GFA version becomes clear, from the
    syntax or type of a line.

    Sometimes it is necessary to call this method, when constructing manually
    Gfa instances, which are not complete.
    """
    if self._version is None:
      self._version = self._version_guess
    for i in range(0,len(self._line_queue)):
      self.add_line(self._line_queue[i])
    self._line_queue = []

  def _register_line(self, gfa_line):
    self._api_private_check_gfa_line(gfa_line, "_register_line")
    storage_key = gfa_line.__class__.STORAGE_KEY
    if storage_key == "merge":
      self._records[gfa_line.record_type]._merge(gfa_line)
    elif storage_key == "name":
      if gfa_line.record_type not in self._records:
        self._records[gfa_line.record_type] = {}
      key = gfa_line.name
      if gfapy.is_placeholder(key):
        key = id(gfa_line)
      elif key.isdigit():
        keynum = int(key)
        if keynum > self._max_int_name:
          self._max_int_name = keynum
      self._records[gfa_line.record_type][key] = gfa_line
    elif storage_key == "external":
      if gfa_line.external.line not in self._records[gfa_line.record_type]:
        self._records[gfa_line.record_type][gfa_line.external.line] = {}
      self._records[gfa_line.record_type][\
          gfa_line.external.line][id(gfa_line)] = gfa_line
    elif storage_key is None:
      if gfa_line.record_type not in self._records:
        self._records[gfa_line.record_type] = {}
      self._records[gfa_line.record_type][id(gfa_line)] = gfa_line

  def __add_line_unknown_version(self, gfa_line):
    if isinstance(gfa_line, str):
      rt = gfa_line[0]
    elif isinstance(gfa_line, gfapy.Line):
      rt = gfa_line.record_type
    else:
      raise gfapy.ArgumentError(\
          "Only strings and gfapy.Line instances can be added")
    if rt == "#":
      if isinstance(gfa_line, str):
        gfa_line = gfapy.Line(gfa_line, dialect=self._dialect)
      gfa_line.connect(self)
    elif rt == "H":
      if isinstance(gfa_line, str):
        gfa_line = gfapy.Line(gfa_line, vlevel=self._vlevel,
            dialect=self._dialect)
      self.header._merge(gfa_line)
      if gfa_line.VN:
        if gfa_line.VN == "1.0":
          self._version = "gfa1"
        elif gfa_line.VN == "2.0":
          self._version = "gfa2"
        else:
          self._version = gfa_line.VN
        self._version_explanation = "specified in header VN tag"
        if self._vlevel > 0:
          self._validate_version()
        self.process_line_queue()
    elif rt == "S":
      if isinstance(gfa_line, str):
        gfa_line = gfapy.Line(gfa_line, vlevel=self._vlevel,
            dialect=self._dialect)
      self._version = gfa_line.version
      self._version_explanation = \
          "implied by: syntax of S {} line".format(gfa_line.name)
      self.process_line_queue()
      gfa_line.connect(self)
    elif rt in ["E", "F", "G", "U", "O"]:
      self._version = "gfa2"
      self._version_explanation = "implied by: presence of a {} line".format(rt)
      if isinstance(gfa_line, str):
        gfa_line = gfapy.Line(gfa_line, vlevel=self._vlevel,
            version=self._version, dialect=self._dialect)
      self.process_line_queue()
      gfa_line.connect(self)
    elif rt in ["L", "C", "P"]:
      self._version_guess = "gfa1"
      self._line_queue.append(gfa_line)
    else:
      self._line_queue.append(gfa_line)

  def __add_line_GFA1(self, gfa_line):
    if isinstance(gfa_line, str):
      if gfa_line[0] == "S":
        gfa_line = gfapy.Line(gfa_line, vlevel=self._vlevel,
            dialect=self._dialect)
      else:
        gfa_line = gfapy.Line(gfa_line, vlevel=self._vlevel,
            dialect=self._dialect, version="gfa1")
    elif gfa_line.__class__ in gfapy.Lines.GFA2Specific:
      raise gfapy.VersionError(
        "Version: 1.0 ({})\n".format(self._version_explanation)+
        "Cannot add instance of incompatible line type "+
        str(type(gfa_line)))
    if gfa_line.record_type == "H":
      if self._vlevel > 0 and gfa_line.VN and gfa_line.VN != "1.0":
        raise gfapy.VersionError(
          "Header line specified wrong version ({})\n".format(gfa_line.VN)+
          "Line: {}\n".format(gfa_line)+
          "File version: 1.0 ({})".format(self._version_explanation))
      self.header._merge(gfa_line)
    elif gfa_line.record_type == "S":
      if gfa_line.version == "gfa2":
        raise gfapy.VersionError(
          "Version: 1.0 ({})\n".format(self._version_explanation)+
          "GFA2 segment found: {}".format(gfa_line))
      gfa_line.connect(self)
    elif gfa_line.record_type in ["L", "P", "C", "#"]:
      gfa_line.connect(self)
    else:
      raise gfapy.AssertionError(
        "Invalid record type {}. This should never happen".format(rt))

  def __add_line_GFA2(self, gfa_line):
    if isinstance(gfa_line, str):
      if gfa_line[0] == "S":
        gfa_line = gfapy.Line(gfa_line, vlevel=self._vlevel,
            dialect=self._dialect)
      else:
        gfa_line = gfapy.Line(gfa_line, vlevel=self._vlevel,
                                        version="gfa2", dialect=self._dialect)
    elif gfa_line.__class__ in gfapy.Lines.GFA1Specific:
      raise gfapy.VersionError(
        "Version: 2.0 ({})\n".format(self._version_explanation)+
        "Cannot add instance of incompatible line type "+
        str(type(gfa_line)))
    if gfa_line.record_type == "H":
      if self._vlevel > 0 and gfa_line.VN and gfa_line.VN != "2.0":
        raise gfapy.VersionError(
          "Header line specified wrong version ({})\n".format(gfa_line.VN)+
          "Line: {}\n".format(gfa_line)+
          "File version: 2.0 ({})".format(self._version_explanation))
      self.header._merge(gfa_line)
    elif gfa_line.record_type == "S":
      if gfa_line.version == "gfa1":
        raise gfapy.VersionError(
          "Version: 2.0 ({})\n".format(self._version_explanation)+
          "GFA1 segment found: {}".format(gfa_line))
      gfa_line.connect(self)
    else:
      gfa_line.connect(self)




from .comment import Comment
from .line import Line
from .unknown import Unknown
from .edge import Edge
from .gap import Gap
from .custom_record import CustomRecord
from .fragment import Fragment
from .header import Header
from .segment import Segment
from . import group



import gfapy
import re

def decode(string):
  validate_encoded(string)
  return string

def unsafe_decode(string):
  return string

def validate_encoded(string):
  if not re.match("^[ !-~]+$", string):
    raise gfapy.FormatError(
      "{} is not a valid string field\n".format(repr(string))+
      "(it contains newlines/tabs and/or non-printable characters)")

validate_decoded = validate_encoded

def unsafe_encode(obj):
  return str(obj)

def encode(obj):
  if not isinstance(obj, str):
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str)")
  validate_encoded(obj)
  return obj



import unittest
import gfapy

class TestUnitFieldArray(unittest.TestCase):

  def test_initialize(self):
    a = gfapy.FieldArray("i", [1,2,3])
    # from a FieldArray:
    z = gfapy.FieldArray("Z", a)
    # no validations by default:
    gfapy.FieldArray("i", [1,2,"a"])
    gfapy.FieldArray("wrong", [1,2])

  def test_datatype(self):
    fa = gfapy.FieldArray("i", [1,2,3])
    self.assertEqual("i", fa.datatype)

  def test_validate(self):
    f1 = gfapy.FieldArray("i", [1,2,3])
    f2 = gfapy.FieldArray("i", [1,2,"a"])
    f3 = gfapy.FieldArray("wrong", [1,2])
    f1.validate()
    self.assertRaises(gfapy.FormatError, f2.validate)
    self.assertRaises(gfapy.TypeError, f3.validate)

  def test_validate_gfa_field(self):
    gfapy.FieldArray("i", [1,2,3])._validate_gfa_field("i")
    self.assertRaises(gfapy.TypeError,
        gfapy.FieldArray("i", [1,2,3])._validate_gfa_field, "J")
    self.assertRaises(gfapy.FormatError,
        gfapy.FieldArray("i", [1,2,"a"])._validate_gfa_field, "i")
    gfapy.FieldArray("wrong", [1,2])._validate_gfa_field("i")

  def test_to_gfa_field(self):
    f = gfapy.FieldArray("i", [1,2,3])
    self.assertEqual("1\t2\t3", f._to_gfa_field())

  def test_to_gfa_tag(self):
    f = gfapy.FieldArray("i", [1,2,3])
    self.assertEqual("xx:i:1\txx:i:2\txx:i:3", f._to_gfa_tag("xx"))

  def test_vpush(self):
    self.assertRaises(gfapy.FormatError,
      gfapy.FieldArray("i", [1,2,3])._vpush, "x")
    self.assertRaises(gfapy.TypeError,
      gfapy.FieldArray("i", [1,2,3])._vpush, 2.0)
    self.assertRaises(gfapy.InconsistencyError,
      gfapy.FieldArray("i", [1,2,3])._vpush, "z", "Z")
    gfapy.FieldArray("i", [1,2,3])._vpush("z", "i")




import unittest
import gfapy

class TestUnitLineCloning(unittest.TestCase):

  def test_clone_tags(self):
    l = gfapy.Line("H\tVN:Z:1.0")
    l1 = l
    l2 = l.clone()
    self.assertIsInstance(l, gfapy.line.Header)
    self.assertIsInstance(l2, gfapy.line.Header)
    l2.VN = "2.0"
    self.assertEqual("2.0", l2.VN)
    self.assertEqual("1.0", l.VN)
    l1.VN = "2.0"
    self.assertEqual("2.0", l.VN)

  def test_clone_deep_string(self):
    s = gfapy.Line("S\t1\tCAGCTTG")
    s_clone = s.clone()
    s_clone.sequence += "CCC"
    self.assertNotEqual(s_clone.sequence, s.sequence)

  def test_clone_deep_posfield_array(self):
    u = gfapy.Line("U\t*\t1 2 3")
    u_clone = u.clone()
    self.assertEqual(u_clone.items, u.items)
    self.assertNotEqual(id(u_clone.items), id(u.items))
    u_clone.items.append("4")
    self.assertNotEqual(u_clone.items, u.items)

  def test_clone_deep_J_field(self):
    h = gfapy.Line("H\txx:J:[1,2,3]")
    h_clone = h.clone()
    self.assertEqual(h_clone.xx, h.xx)
    self.assertNotEqual(id(h_clone.xx), id(h.xx))
    h_clone.xx[0] += 1
    self.assertNotEqual(h_clone.xx, h.xx)

  def test_clone_disconnected(self):
    g = gfapy.Gfa()
    sA = gfapy.Line("S\tA\t7\tCAGCTTG")
    u = gfapy.Line("U\tU12\tA B C")
    g.add_line(sA)
    g.add_line(u)
    assert(u.is_connected())
    self.assertEqual([u], sA.sets)
    self.assertEqual([u], g.sets)
    u_clone = u.clone()
    assert(not u_clone.is_connected())
    self.assertEqual([u], sA.sets)
    self.assertEqual([u], g.sets)
    assert(all(isinstance(i,gfapy.Line) for i in u.items))
    assert(not any(isinstance(i,gfapy.Line) for i in u_clone.items))
    self.assertEqual(["A", "B", "C"], [e.name for e in u.items])
    self.assertEqual(["A", "B", "C"], u_clone.items)



import gfapy
import unittest

class TestAPILinesCreators(unittest.TestCase):

  def test_add_headers(self):
    gfa = gfapy.Gfa()
    h = "H\tVN:Z:1.0"
    gfa.append(h) # nothing raised
    self.assertEqual([h], [str(x) for x in gfa.headers])

  def test_add_segments(self):
    gfa = gfapy.Gfa()
    s1 = gfapy.Line("S\t1\t*")
    s2 = gfapy.Line("S\t2\t*")
    s2c = gfapy.Line("S\t2\t*")
    gfa.append(s1) # nothing raised
    gfa.append(s2) # nothing raised
    self.assertSetEqual(set([str(s) for s in [s1, s2]]), set([str(s) for s in gfa.segments]))
    self.assertSetEqual(set(["1", "2"]), set(gfa.segment_names))
    self.assertEqual(s1, gfa.segment("1"))
    self.assertEqual(None, gfa.segment("0"))
    gfa.try_get_segment("1") # nothing raised
    with self.assertRaises(gfapy.NotFoundError): gfa.try_get_segment("0")
    with self.assertRaises(gfapy.NotUniqueError): gfa.append(s2c)

  def test_add_links(self):
    s1 = "S\t1\t*"
    s2 = "S\t2\t*"
    l1 = gfapy.Line("L\t1\t+\t2\t+\t12M")
    l2 = "L\t1\t+\t3\t+\t12M"
    gfa = gfapy.Gfa()
    gfa.append(s1)
    gfa.append(s2)
    gfa.append(l1) # nothing raised
    self.assertEqual([l1], gfa.dovetails)
    self.assertEqual([l1], gfa.segment("1").end_relations("R", ["2", "L"]))
    self.assertEqual([l1], gfa.segment("2").end_relations("L", ["1", "R"]))
    self.assertEqual([], gfa.segment("2").end_relations("R", gfapy.SegmentEnd("1", "L")))
    gfa.append(l2) # nothing raised

  def test_add_containments(self):
    s1 = "S\t1\t*"
    s2 = "S\t2\t*"
    c1 = gfapy.Line("C\t1\t+\t2\t+\t12\t12M")
    c2 = "C\t1\t+\t3\t+\t12\t12M"
    gfa = gfapy.Gfa()
    gfa.append(s1)
    gfa.append(s2)
    gfa.append(c1) # nothing raised
    self.assertEqual([c1], gfa.containments)
    self.assertEqual([c1],
             gfa.segment("1").relations_to("2", "edges_to_contained"))
    self.assertEqual([],
             gfa.segment("2").relations_to("1", "edges_to_contained"))
    gfa.append(c2) # nothing raised

  def test_add_paths(self):
    s1 = "S\t1\t*"
    s2 = "S\t2\t*"
    p1 = gfapy.Line("P\t4\t1+,2+\t122M")
    p2 = "P\t1\t1+,2+\t122M"
    p3 = "P\t5\t1+,2+,3+\t122M,120M"
    gfa = gfapy.Gfa()
    gfa.append(s1)
    gfa.append(s2)
    gfa.append(p1) # nothing raised
    self.assertEqual([p1], gfa.paths)
    self.assertEqual(["4"], gfa.path_names)
    self.assertEqual(p1, gfa.line("4"))
    self.assertEqual(None, gfa.line("5"))
    with self.assertRaises(gfapy.NotUniqueError): gfa.append(p2)
    gfa.append(p3) # nothing raised

##  def test_segments_first_order(self):
##    s1 = "S\t1\t*"
##    s2 = "S\t2\t*"
##    l1 = "L\t1\t+\t2\t+\t122M"
##    l2 = "L\t1\t+\t3\t+\t122M"
##    c1 = "C\t1\t+\t2\t+\t12\t12M"
##    c2 = "C\t1\t+\t3\t+\t12\t12M"
##    p1 = "P\t4\t1+,2+\t122M"
##    p2 = "P\t1\t1+,2+\t122M"
##    p3 = "P\t5\t1+,3+\t122M"
##    gfa = gfapy.Gfa()
##    gfa.append(s1)
##    gfa.append(s2)
##    gfa.append(l1) # nothing raised
##    with self.assertRaises(gfapy.NotFoundError): gfa.append(l2)
##    gfa.append(c1) # nothing raised
##    with self.assertRaises(gfapy.NotFoundError): gfa.append(c2)
##    gfa.append(p1) # nothing raised
##    with self.assertRaises(gfapy.NotUniqueError): gfa.append(p2)
##    with self.assertRaises(gfapy.NotFoundError): gfa.append(p3)

  def test_header_add(self):
    gfa = gfapy.Gfa()
    gfa.append("H\tVN:Z:1.0")
    gfa.append("H\taa:i:12\tab:Z:test1")
    gfa.append("H\tac:Z:test2")
    gfa.header.add("aa", 15)
    self.assertSetEqual(
      set([
        "H\tVN:Z:1.0",
        "H\taa:i:12",
        "H\taa:i:15",
        "H\tab:Z:test1",
        "H\tac:Z:test2",
      ]),
      set([str(x) for x in gfa.headers]))
    gfa.header.add("aa", 16)
    self.assertSetEqual(
      set([
        "H\tVN:Z:1.0",
        "H\taa:i:12",
        "H\taa:i:15",
        "H\taa:i:16",
        "H\tab:Z:test1",
        "H\tac:Z:test2",
      ]),
      set([str(x) for x in gfa.headers]))
    gfa.header.delete("aa")
    gfa.header.aa = 26
    self.assertEqual(
      set([
        "H\tVN:Z:1.0",
        "H\taa:i:26",
        "H\tab:Z:test1",
        "H\tac:Z:test2",
      ]),
      set([str(x) for x in gfa.headers]))




import unittest
import gfapy

class TestByteArray(unittest.TestCase):

  def test_byte_arrays(self):
    # creation: from list, from string
    a_lst = [18, 172, 244, 170, 96, 28, 31]
    a = gfapy.ByteArray(a_lst)
    for i in range(0,len(a_lst)):
      self.assertEqual(a[i], a_lst[i])
    a_str = "12ACF4AA601C1F"
    b = gfapy.ByteArray(a_str)
    self.assertEqual(a,b)
    # validation
    self.assertRaises(gfapy.ValueError, gfapy.ByteArray, [1,2,3,4,356])
    self.assertRaises(gfapy.FormatError, gfapy.ByteArray, "12ACF4AA601C1")
    self.assertRaises(gfapy.FormatError, gfapy.ByteArray, "")
    self.assertRaises(gfapy.FormatError, gfapy.ByteArray, "12ACG4AA601C1")
    # to_s
    self.assertEqual(str(b), a_str)
    self.assertEqual(str(a), a_str)
    # read only; transform to list to edit a value
    tmp = list(a)
    tmp[3]=1
    a = gfapy.ByteArray(tmp)
    self.assertEqual(a, gfapy.ByteArray([18,172,244,1,96,28,31]))




import gfapy
import re

class OrientedLine:
  """A line plus an orientation.

  The line can be an instance of `~gfapy.line.line.Line` or a string (line
  identifier). The orientation is a string, either ``'+'`` or ``'-'``.
  Methods not defined in this class are delegated to the line element.

  Parameters:
    value (str, list, OrientedLine) : a line identifier with a 1-letter
      orientation suffix + or -, or a list of two elements (identifier
      or line instance and orientation string), or an OrientedLine instance

  Returns:
    OrientedLine: if value is an OrientedLine, then
      it is returned; if it is a string, then an OrientedLine where line
      is a string (the string without the last char, which is the orientation);
      if it is a list, then an OrientedLine where line is the first element,
      orientation the second
  """

  def __new__(cls, *args):
    if isinstance(args[0], OrientedLine):
      return args[0]
    else:
      new_instance = object.__new__(cls)
      return new_instance

  def __init__(self, *args):
    if len(args) == 1:
      if isinstance(args[0], OrientedLine):
        return
      elif isinstance(args[0], str):
        self.__line = args[0][0:-1]
        self.__orient = args[0][-1]
      elif isinstance(args[0], list):
        self.__line = args[0][0]
        self.__orient = args[0][1]
      else:
        raise gfapy.ArgumentError("Cannot create an OrientedLine"+
            " instance from an object of type {}".format(type(args[0])))
    elif len(args) == 2:
      self.__line = args[0]
      self.__orient = args[1]
    else:
      raise gfapy.ArgumentError("Wrong number of arguments for OrientedLine()")
    self.__editable = True

  @property
  def line(self):
    """The line.

    Returns:
      str or `~gfapy.line.line.Line`
    """
    return self.__line

  @property
  def orient(self):
    """The orientation.

    Returns:
      str : '+' or '-'
    """
    return self.__orient

  @line.setter
  def line(self, line):
    if self.__editable:
      self.__line = line
    else:
      raise gfapy.RuntimeError(
          "gfapy.OrientedLine instance cannot be edited ({})".format(self))

  @orient.setter
  def orient(self, orient):
    if self.__editable:
      self.__orient = orient
    else:
      raise gfapy.RuntimeError(
          "gfapy.OrientedLine instance cannot be edited ({})".format(self))

  @property
  def name(self):
    """The name of the line.

    Returns:
      str : if line is a string, then line; if it is a line instance,
            then line.name
    """
    if isinstance(self.__line, str):
      return self.__line
    else:
      return self.__line.name

  def validate(self):
    """Validate the content of the instance

    Raises:
      gfapy.error.ValueError: if the orientation is invalid
      gfapy.error.TypeError: if the line is not a string or a
        `gfapy.line.line.Line` instance
      gfapy.error.FormatError: if the line is a string which is not a valid
        line identifier, or it is a Line instance with an invalid name
    """
    self.__validate_line()
    self.__validate_orient()
    return None

  def inverted(self):
    """An oriented line with the same line element, but inverted orientation.

    Note:
      the inverted() method returns an OrientedLine with inverted orientation;
      the invert() method inverts the orientation in place (and returns None)
    """
    return OrientedLine(self.line, gfapy.invert(self.orient))

  def invert(self):
    """Invert the orientation of the OrientedLine instance.

    Note:
      the inverted() method returns an OrientedLine with inverted orientation;
      the invert() method inverts the orientation in place (and returns None)
    """
    self.orient = gfapy.invert(self.orient)

  def __str__(self):
    if self.name:
      return "{}{}".format(self.name, self.orient)
    else:
      return "({}){}".format(str(self.line), self.orient)

  def __repr__(self):
    return "gfapy.OrientedLine({},{})".format(repr(self.line),repr(self.orient))

  def __eq__(self, other):
    if isinstance(other, OrientedLine):
      pass
    elif isinstance(other, list):
      other = OrientedLine(other)
    elif isinstance(other, str):
      other = OrientedLine(other)
    else:
      return False
    return (self.name == other.name) and (self.orient == other.orient)

  # Delegate methods to the line
  def __getattr__(self, name):
    return getattr(self.__line, name)

  def _block(self):
    self.__editable = False

  def _unblock(self):
    self.__editable = True

  def __validate_orient(self):
    if not self.orient in ["+", "-"]:
      raise gfapy.ValueError("Invalid orientation ({})".format(self.orient))

  def __validate_line(self):
    if isinstance(self.line, gfapy.Line):
      string = self.line.name
    elif isinstance(self.line, str):
      string = self.line
    else:
      raise gfapy.TypeError(
        "Invalid class ({}) for line reference ({})"
        .format(self.line.__class__, self.line))
    if not re.match(r"^[!-~]+$", string):
      raise gfapy.FormatError(
      "{} is not a valid GFA identifier\n".format(repr(string))+
      "(it contains spaces or non-printable characters)")



import unittest
import gfapy

class TestTrace(unittest.TestCase):
  def test_from_string(self):
    self.assertEqual(gfapy.Trace([12,14,15]), gfapy.Trace._from_string("12,14,15"))
    with self.assertRaises(gfapy.FormatError):
      gfapy.Trace._from_string("12x,12,12")

  def test_validation(self):
    gfapy.Trace._from_string("12,12,12").validate()
    self.assertRaises(gfapy.ValueError, gfapy.Trace._from_string("12,12,12").validate, ts = 10)
    self.assertRaises(gfapy.ValueError, gfapy.Trace._from_string("12,-12,12").validate, ())
    self.assertRaises(gfapy.TypeError, gfapy.Trace(["12x",12,12]).validate, ())

  def test_str(self):
    self.assertEqual("12,12,12", str(gfapy.Trace([12,12,12])))



import gfapy
import re

class SegmentEnd:
  """A segment plus an end type (L or R).

  The ``segment`` can be an instance of a segment subclass of
  `~gfapy.line.line.Line` or a string (line identifier). The ``end_type``
  symbol is a string, either ``'L'`` or ``'R'``. Methods not defined in this
  class are delegated to the segment element.

  Parameters:
    value (str, list, SegmentEnd) : a line identifier with a 1-letter
      end symbol L or R, or a list of two elements (identifier
      or line instance and end symbol), or an SegmentEnd instance

  Returns:
    SegmentEnd: if value is an SegmentEnd, then
      it is returned; if it is a string, then an SegmentEnd where line
      is a string (the string without the last char, which is the end symbol);
      if it is a list, then an SegmentEnd where line is the first element,
      end symbol the second
  """

  def __new__(cls, *args):
    if isinstance(args[0], SegmentEnd):
      return args[0]
    else:
      new_instance = object.__new__(cls)
      return new_instance

  def __init__(self, *args):
    if len(args) == 1:
      if isinstance(args[0], SegmentEnd):
        return
      elif isinstance(args[0], str):
        self.__segment = args[0][0:-1]
        self.__end_type = args[0][-1]
      elif isinstance(args[0], list):
        if len(args[0]) != 2:
          raise gfapy.ArgumentError("Cannot create a SegmentEnd "+
            " from a list of size {}".format(len(args[0])))
        self.__segment = args[0][0]
        self.__end_type = args[0][1]
      else:
        raise gfapy.ArgumentError("Cannot create an SegmentEnd "+
            " from an object of type {}".format(type(args[0])))
    elif len(args) == 2:
      self.__segment = args[0]
      self.__end_type = args[1]
    else:
      raise gfapy.ArgumentError("Wrong number of arguments for SegmentEnd()")

  def validate(self):
    """Validate the content of the instance

    Raises:
      gfapy.error.ValueError: if the orientation is invalid
      gfapy.error.TypeError: if the segment is not a string or
        an instance of a segment subclass of `gfapy.line.line.Line`
      gfapy.error.FormatError: if the segment is a string which is not a valid
        segment identifier, or it is a segment Line instance with an invalid
        name
    """
    self.__validate_segment()
    self.__validate_end_type()
    return None

  def __validate_end_type(self):
    if not self.__end_type in ["L", "R"]:
      raise gfapy.ValueError(
          "Invalid end type ({})".format(repr(self.__end_type)))

  def __validate_segment(self):
    if isinstance(self.segment, gfapy.line.Segment):
      string = self.segment.name
    elif isinstance(self.segment, str):
      string = self.segment
    else:
      raise gfapy.TypeError(
        "Invalid class ({}) for segment reference ({})"
        .format(self.segment.__class__, self.segment))
    if not re.match(r"^[!-~]+$", string):
      raise gfapy.FormatError(
      "{} is not a valid segment identifier\n".format(repr(string))+
      "(it contains spaces or non-printable characters)")

  @property
  def segment(self):
    """The segment.

    Returns:
      str or `gfapy.line.segment.gfa1.GFA1` or `gfapy.line.segment.gfa2.GFA2`
    """
    return self.__segment

  @segment.setter
  def segment(self, value):
    self.__segment=value

  @property
  def name(self):
    """The name of the segment.

    Returns:
      str : if segment is a string, then segment; if it is a segment instance,
            then segment.name
    """
    if isinstance(self.__segment, gfapy.Line):
      return self.__segment.name
    else:
      return str(self.__segment)

  @property
  def end_type(self):
    """The end type.

    Returns:
      str : 'L' or 'R'
    """
    return self.__end_type

  @end_type.setter
  def end_type(self, value):
    self.__end_type = value

  def inverted(self):
    return SegmentEnd(self.__segment, gfapy.invert(self.end_type))

  def __str__(self):
    return "{}{}".format(self.name, self.end_type)

  def __repr__(self):
    return "gfapy.SegmentEnd({},{})".format(repr(self.segment),
                                            repr(self.end_type))

  def __eq__(self, other):
    if isinstance(other, list):
      other = SegmentEnd(other)
    elif isinstance(other, str):
      other = SegmentEnd(other)
    elif not isinstance(other, gfapy.SegmentEnd):
      return False
    return (self.name == other.name) and (self.end_type == other.end_type)

  def __getattr__(self, name):
    return getattr(self.__segment, name)



import re
import gfapy

class Validate:

  def validate_field(self, fieldname):
    """
    Raises an error if the content of the field does not correspond to
    the field type.

    Parameters
    ----------
    fieldname : str
      The tag name of the field to validate.

    Raises
    ------
    gfapy.FormatError
      If the content of the field is not valid, according to its required type.
    """
    fieldname = self.__class__.FIELD_ALIAS.get(fieldname, fieldname)
    v = self._data[fieldname]
    t = self._field_or_default_datatype(fieldname, v)
    gfapy.Field._validate_gfa_field(v, t, fieldname)

  def validate(self):
    """
    Validate the gfapy.Line instance.

    Raises
    ------
    gfapy.FormatError
      If any field content is not valid.
    """
    fieldnames = self.positional_fieldnames + self.tagnames
    if self.vlevel == 0:
      self._validate_tagnames_and_types()
    for fieldname in fieldnames:
      self.validate_field(fieldname)
    self._validate_record_type_specific_info()

  def _validate_tagnames_and_types(self):
    for n in self.tagnames:
      if self._is_predefined_tag(n):
        self._validate_predefined_tag_type(n, self._field_datatype(n))
      elif not self._is_valid_custom_tagname(n):
        raise gfapy.FormatError("Custom tag names must consist in a letter "+
            "and a digit or two letters\nFound: {}".format(tagname))

  def _validate_predefined_tag_type(self, tagname, datatype):
    if datatype != self.__class__.DATATYPE[tagname]:
      raise gfapy.TypeError(
        "Tag {} must be of type ".format(tagname) +
        "{}, {} found".format(self.__class__.DATATYPE[tagname], datatype))

  def _validate_custom_tagname(self, tagname):
    if not self._is_valid_custom_tagname(tagname):
      raise gfapy.FormatError("Custom tag names must consist in a letter "+
          "and a digit or two letters\nFound: {}".format(tagname))

  @staticmethod
  def _is_valid_custom_tagname(tagname):
    return (re.match(r"^[A-Za-z][A-Za-z0-9]$", tagname))

  def _validate_record_type_specific_info(self):
    pass

  def _is_predefined_tag(self, fieldname):
    return fieldname in self.__class__.PREDEFINED_TAGS




import gfapy
import unittest

class TestAPIGfaBasics(unittest.TestCase):

  def test_initialize(self):
    gfapy.Gfa() # nothing raised
    gfa = gfapy.Gfa()
    self.assertEqual(gfapy.Gfa, gfa.__class__)

  def test_version_empty(self):
    gfa = gfapy.Gfa()
    self.assertIsNone(gfa.version)
    gfa = gfapy.Gfa(version="gfa1")
    self.assertEqual("gfa1", gfa.version)
    gfa = gfapy.Gfa(version="gfa2")
    self.assertEqual("gfa2", gfa.version)
    with self.assertRaises(gfapy.VersionError): gfapy.Gfa(version="0.0")

  def test_validate(self):
    gfa = gfapy.Gfa(version="gfa1")
    gfa.append("S\t1\t*")
    gfa.validate() # nothing raised
    gfa.append("L\t1\t+\t2\t-\t*")
    with self.assertRaises(gfapy.NotFoundError): gfa.validate()
    gfa.append("S\t2\t*")
    gfa.validate() # nothing raised
    gfa.append("P\t3\t1+,4-\t*")
    with self.assertRaises(gfapy.NotFoundError): gfa.validate()
    gfa.append("S\t4\t*")
    with self.assertRaises(gfapy.NotFoundError): gfa.validate()
    gfa.append("L\t4\t+\t1\t-\t*")
    gfa.validate() # nothing raised

  def test_to_s(self):
    lines = ["H\tVN:Z:1.0","S\t1\t*","S\t2\t*","S\t3\t*",
     "L\t1\t+\t2\t-\t*","C\t1\t+\t3\t-\t12\t*","P\t4\t1+,2-\t*"]
    gfa = gfapy.Gfa()
    for l in lines: gfa.append(l)
    self.assertEqual(set(lines), set(str(gfa).split("\n")))

  ## def test_from_file(self):
  ##   filename = "tests/testdata/example1.gfa"
  ##   gfa = gfapy.Gfa.from_file(filename)
  ##   assert(gfa)
  ##   with open(filename) as f:
  ##     txt = f.read()
  ##   self.assertEqual(txt, str(gfa))

  ## def test_to_file(self):
  ##   filename = "tests/testdata/example1.gfa"
  ##   gfa = gfapy.Gfa.from_file(filename)
  ##   tmp = Tempfile("example1")
  ##   gfa.to_file(tmp.path)
  ##   tmp.rewind
  ##   self.assertEqual(IO.read(filename), IO.read(tmp))

  def test_from_string(self):
    lines = ["H\tVN:Z:1.0","S\t1\t*","S\t2\t*","S\t3\t*",
     "L\t1\t+\t2\t-\t*","C\t1\t+\t3\t-\t12\t*","P\t4\t1+,2-\t*"]
    gfa1 = gfapy.Gfa()
    for l in lines: gfa1.append(l)
    gfa2 = gfapy.Gfa("\n".join(lines))
    assert(gfa2)
    self.assertEqual(gfapy.Gfa, gfa2.__class__)
    self.assertEqual(str(gfa1), str(gfa2))

  def test_from_list(self):
    lines = ["H\tVN:Z:1.0","S\t1\t*","S\t2\t*","S\t3\t*",
     "L\t1\t+\t2\t-\t*","C\t1\t+\t3\t-\t12\t*","P\t4\t1+,2-\t*"]
    gfa1 = gfapy.Gfa()
    for l in lines: gfa1.append(l)
    gfa2 = gfapy.Gfa(lines)
    assert(gfa2)
    self.assertEqual(gfapy.Gfa, gfa2.__class__)
    self.assertEqual(str(gfa1), str(gfa2))




import gfapy

class Writer:

  def __str__(self):
    """
    Returns
    -------
    str
      A string representation of self.
    """
    return gfapy.Line.SEPARATOR.join(self.to_list())

  def to_str(self, add_virtual_commentary=True):
    """
    Parameters
    ----------
    add_virtual_commentary : bool
      add a 'co' tag to virtual lines (default: True)

    Returns
    -------
    str
      A string representation of self.
    """
    return gfapy.Line.SEPARATOR.join(self.to_list(
      add_virtual_commentary=add_virtual_commentary))

  def to_list(self, add_virtual_commentary=True):
    """
    Parameters
    ----------
    add_virtual_commentary : bool
      add a 'co' tag to virtual lines (default: True)

    Returns
    -------
    str list
      A list of string representations of the fields.
    """
    a = []
    errors = []
    try:
      rt = self.record_type
    except:
      rt = "<error>"
      errors.append("record_type")
    a.append(rt)
    for fn in self.positional_fieldnames:
      try:
        fstr = self.field_to_s(fn, tag = False)
      except:
        fstr = str(self.get(fn))
        errors.append(fn)
      a.append(fstr)
    for fn in self.tagnames:
      try:
        fstr = self.field_to_s(fn, tag = True)
      except:
        fstr = str(self.get(fn))
        errors.append(fn)
      a.append(fstr)
    if self.virtual and add_virtual_commentary:
      a.append("co:Z:GFAPY_virtual_line")
    if errors:
      a.append("# INVALID; errors found in fields: "+
          ",".join(errors))
    return a

  def field_to_s(self, fieldname, tag = False):
    """
    Compute the string representation of a field.

    Parameters
    ----------
    fieldname : str
      The tag name of the field.
    tag : bool
      *(defaults to: ***False***)*
      Return the tagname:datatype:value representation.

    Raises
    ------
    gfapy.NotFoundError
      If field is not defined.

    Returns
    -------
    str
      The string representation
    """
    fieldname = self.__class__.FIELD_ALIAS.get(fieldname, fieldname)
    v = self._data.get(fieldname, None)
    if v is None:
      raise gfapy.NotFoundError("Field {} not found".format(fieldname))
    t = self._field_or_default_datatype(fieldname, v)
    if not isinstance(v, str):
      v = gfapy.Field._to_gfa_field(v, datatype = t, fieldname = fieldname,
                                  line = self)
    if self.vlevel >= 2:
      gfapy.Field._validate_gfa_field(v, t, fieldname)
    if tag:
      return gfapy.Field._to_gfa_tag(v, fieldname, datatype = t, line = self)
    else:
      return v

  def __repr__(self):
    try:
      s = str(self)
    except:
      rt = self.record_type + "(error!)"
      s = [ rt ]
      for fn in self.positional_fieldnames:
        try:
          field_s = repr(self.get(fn))
        except:
          field_s = "<error>"
        s.append(field_s)
      for tn in self.tagnames:
        dt = self.get_datatype(tn)
        try:
          tv = repr(self.get(tn))
        except:
          tv = "<ERROR>"
        s.append("{}:{}:{}".format(tn,dt,tv))
      s = "\t".join(s)
    return "gfapy.Line('{0}',version='{1}',vlevel={2})".format(s,self.version,self.vlevel)

  def refstr(self, maxlen=10):
    """String containing a list of lines referencing to this line.

    Parameters
    ----------
    maxlen : int
      Shorten lists longer than the specified value (default: 10)

    Returns
    -------
    str
    """
    andmore = 0
    references = self.all_references
    if len(references) > maxlen:
      andmore = len(references) - 10
      references = references[:10]
    lines_list = "\n".join([str(l) for l in references])
    if andmore > 0:
      lines_list += "\n... ({} more)".format(andmore)
    return lines_list

  @property
  def _tags(self):
    """
    Returns the tags as an array of [fieldname, datatype, value]
    triples.

    Returns
    -------
    (str, str, object) list
    """
    retval = []
    for of in tagnames:
      retval.append([of, self.get_datatype(of), self.get(of)])
    return retval



import gfapy

class Headers:
  @property
  def header(self):
    """The header of the Gfa instance.

    For simplicity of access, all tags are summarized in a single
    Header instance. If the same tag is defined on different H lines,
    the values are collected into a FieldArray instance.
    """
    return self._records["H"]

  @property
  def headers(self):
    """The splitted header of the Gfa instance.

    The header of the Gfa instance, splitted into H lines containing
    each a single tag.
    """
    return self._records["H"]._split()



import gfapy
import unittest

class TestInternalsFieldWriter(unittest.TestCase):
  def test_field_writer_i(self):
    self.assertEqual("13", gfapy.Field._to_gfa_field(13))

  def test_field_writer_f(self):
    self.assertEqual("1.3", gfapy.Field._to_gfa_field(1.3))

  def test_field_writer_Z(self):
    self.assertEqual("1B", gfapy.Field._to_gfa_field("1B"))

  def test_field_writer_H(self):
    self.assertEqual("0D0D0D",
                      gfapy.Field._to_gfa_field(gfapy.ByteArray([13,13,13])))
    with self.assertRaises(gfapy.ValueError):
      gfapy.Field._to_gfa_field(gfapy.ByteArray([13,13,1.3]))
    with self.assertRaises(gfapy.ValueError):
      gfapy.Field._to_gfa_field(gfapy.ByteArray([13,13,350]))

  def test_field_writer_B(self):
    self.assertEqual("C,13,13,13", gfapy.Field._to_gfa_field([13,13,13]))
    self.assertEqual("f,1.3,1.3,1.3", gfapy.Field._to_gfa_field([1.3,1.3,1.3]))
    with self.assertRaises(gfapy.ValueError):
      gfapy.Field._to_gfa_field([13,1.3,1.3], "B")

  def test_field_writer_J(self):
    self.assertEqual("[\"A\", 12]", gfapy.Field._to_gfa_field(["A", 12]))
    self.assertEqual("{\"A\": 12}", gfapy.Field._to_gfa_field({"A" : 12}))

  def test_field_writer_as_tag(self):
    self.assertEqual("AA:i:13", gfapy.Field._to_gfa_tag(13, "AA"))



import gfapy

class Collections:
  @property
  def comments(self):
    """List of the comment lines (lines starting with #).

    Note:
      adding or removing elements to the list, does not add or removes
      lines from the Gfa instance. For this, the add_line() and rm() methods
      shall be used. Calling disconnect() on a line of the list, however,
      removes the line from the instance.
    """
    d = self._records["#"]
    return list(d.values())

  @property
  def gaps(self):
    """List of the gap (G) lines. The list is empty in GFA1.

    Note:
      adding or removing elements to the list, does not add or removes
      lines from the Gfa instance. For this, the add_line() and rm() methods
      shall be used. Calling disconnect() on a line of the list, however,
      removes the line from the instance.
    """
    d = self._records["G"]
    return list(d.values())

  @property
  def sets(self):
    """List of the set (U) lines. The list is empty in GFA1.

    Note:
      adding or removing elements to the list, does not add or removes
      lines from the Gfa instance. For this, the add_line() and rm() methods
      shall be used. Calling disconnect() on a line of the list, however,
      removes the line from the instance.
    """
    d = self._records["U"]
    return list(d.values())

  @property
  def segments(self):
    """List of the segment (S) lines.

    Note:
      adding or removing elements to the list, does not add or removes
      lines from the Gfa instance. For this, the add_line() and rm() methods
      shall be used. Calling disconnect() on a line of the list, however,
      removes the line from the instance.
    """
    d = self._records["S"]
    return list(d.values())

  @property
  def edges(self):
    """List of the edge lines.

    Edge lines are L and C lines in GFA1 and E lines in GFA2.

    Note:
      adding or removing elements to the list, does not add or removes
      lines from the Gfa instance. For this, the add_line() and rm() methods
      shall be used. Calling disconnect() on a line of the list, however,
      removes the line from the instance.
    """
    if self._version == "gfa1":
      return self._gfa1_links + self._gfa1_containments
    elif self._version == "gfa2":
      return self._gfa2_edges
    else:
      return self._gfa1_links + self._gfa1_containments + self._gfa2_edges

  @property
  def dovetails(self):
    """List of the dovetail edge lines.

    Dovetail edge lines are L lines in GFA1 and E lines representing dovetail
    overlaps in GFA2.

    Note:
      adding or removing elements to the list, does not add or removes
      lines from the Gfa instance. For this, the add_line() and rm() methods
      shall be used. Calling disconnect() on a line of the list, however,
      removes the line from the instance.
    """
    if self._version == "gfa1":
      return self._gfa1_links
    elif self._version == "gfa2":
      return [ e for e in self._gfa2_edges if e.is_dovetail() ]
    else:
      return self._gfa1_links + \
        [ e for e in self._gfa2_edges if e.is_dovetail() ]

  @property
  def containments(self):
    """List of the containment edge lines.

    Containment edge lines are C lines in GFA1 and E lines representing
    containments in GFA2.

    Note:
      adding or removing elements to the list, does not add or removes
      lines from the Gfa instance. For this, the add_line() and rm() methods
      shall be used. Calling disconnect() on a line of the list, however,
      removes the line from the instance.
    """
    if self._version == "gfa1":
      return self._gfa1_containments
    elif self._version == "gfa2":
      return [ e for e in self._gfa2_edges if e.is_containment() ]
    else:
      return self._gfa1_containments + \
        [ e for e in self._gfa2_edges if e.is_containment() ]

  @property
  def paths(self):
    """List of the path lines (P in GFA1, O in GFA2).

    Note:
      adding or removing elements to the list, does not add or removes
      lines from the Gfa instance. For this, the add_line() and rm() methods
      shall be used. Calling disconnect() on a line of the list, however,
      removes the line from the instance.
    """
    return self._gfa1_paths + self._gfa2_paths

  @property
  def fragments(self):
    """List of the fragment (F) lines. The list is empty in GFA1.

    Note:
      adding or removing elements to the list, does not add or removes
      lines from the Gfa instance. For this, the add_line() and rm() methods
      shall be used. Calling disconnect() on a line of the list, however,
      removes the line from the instance.
    """
    d = self._records["F"]
    return [f for e in d.values() for f in e.values()]

  @property
  def custom_records(self):
    """List of the custom records of the GFA. The list is empty for GFA1.

    All records with a non-standard first field (record type) are considered
    custom records.

    Note:
      adding or removing elements to the list, does not add or removes
      lines from the Gfa instance. For this, the add_line() and rm() methods
      shall be used. Calling disconnect() on a line of the list, however,
      removes the line from the instance.
    """
    cr = []
    for k in self.custom_record_keys:
      collection = self._records[k]
      cr += list(collection.values())
    return cr

  @property
  def _gfa1_containments(self):
    d = self._records["C"]
    return list(d.values())

  @property
  def _gfa1_links(self):
    d = self._records["L"]
    return list(d.values())

  @property
  def _gfa2_edges(self):
    d = self._records["E"]
    return list(d.values())

  @property
  def _gfa2_paths(self):
    d = self._records["O"]
    return list(d.values())

  @property
  def gap_names(self):
    """List of the names of the gap (G) lines. The list is empty in GFA1.
    """
    d = self._records["G"]
    return list([k for k in d.keys() if isinstance(k, str)])

  @property
  def set_names(self):
    """List of the names of the set (U) lines. The list is empty in GFA1.
    """
    d = self._records["U"]
    return list([k for k in d.keys() if isinstance(k, str)])

  @property
  def segment_names(self):
    """List of the names of the segment (S) lines.
    """
    d = self._records["S"]
    return list(d.keys())

  @property
  def edge_names(self):
    """List of the names of the edge (E, L, C) lines.

    For the L and C lines, the content of the custom tag id
    is taken as name.
    """
    if self._version == "gfa1":
      return self._link_names + self._containment_names
    elif self._version == "gfa2":
      return self._gfa2_edge_names
    else:
      return self._gfa2_edge_names + self._link_names + self._containment_names

  @property
  def path_names(self):
    """List of the names of the path lines (P for GFA1, O for GFA2).
    """
    return self._gfa1_path_names + self._gfa2_path_names

  @property
  def names(self):
    """All identifiers in the GFA identifiers namespace.

    Notes:
      GFA1: in Gfapy the P and S namespaces are joined (i.e. paths with
      the same name as segments are not accepted). Furthermore, to simplify
      the conversion to/from GFA2, the ID tag is used in L and C lines,
      and their content is also included in the same namespace as the S/P
      identifiers. GFA2: the namespace for identifiers is described in
      the specification and includes all the S, E, G, U and O lines; the
      external sequence identifiers in F lines are not included.
    """
    return self.segment_names + \
      self.edge_names + \
      self.gap_names + \
      self.path_names + \
      self.set_names

  def unused_name(self):
    """Compute a GFA identifier not yet in use in the Gfa object."""
    self._max_int_name += 1
    return str(self._max_int_name)

  @property
  def external_names(self):
    """List of the identifiers of external sequences mentioned in F records.
    The list is empty in GFA1.
    """
    return list(self._records["F"].keys())

  @property
  def _gfa2_edge_names(self):
    d = self._records["E"]
    return list([k for k in d.keys() if isinstance(k, str)])

  @property
  def _link_names(self):
    d = self._records["L"]
    return list([k for k in d.keys() if isinstance(k, str)])

  @property
  def _containment_names(self):
    d = self._records["C"]
    return list([k for k in d.keys() if isinstance(k, str)])

  @property
  def _gfa2_path_names(self):
    d = self._records["O"]
    return list([k for k in d.keys() if isinstance(k, str)])

  @property
  def _gfa1_paths(self):
    d = self._records["P"]
    return list(d.values())

  @property
  def _gfa1_path_names(self):
    d = self._records["P"]
    return list(d.keys())

  GFA1_ONLY_KEYS = ["L", "C", "P"]

  NONCUSTOM_GFA2_KEYS = ["H", "#", "F", "S", "E", "G", "U", "O", "\t"]

  @property
  def custom_record_keys(self):
    """Record types of the custom records.

    Returns:
      list of str
    """
    if self._version == "gfa1":
      return []
    else:
      keys = [k for k,v in self._records.items() if v]
      if self._version == "gfa2":
        return [k for k in keys if k not in self.NONCUSTOM_GFA2_KEYS]
      else:
        return [k for k in keys \
          if k not in self.NONCUSTOM_GFA2_KEYS and k not in self.GFA1_ONLY_KEYS]

  def custom_records_of_type(self, record_type):
    """List of custom records of the specified type."""
    if record_type not in self.custom_record_keys:
      return []
    return list(self._records[record_type].values())

  @property
  def lines(self):
    """All the lines of the GFA"""
    return self.comments + \
      self.headers + \
      self.segments + \
      self.edges + \
      self.paths + \
      self.sets + \
      self.gaps + \
      self.fragments + \
      self.custom_records



import gfapy
import unittest

class TestUnitMultiplication(unittest.TestCase):

  def test_auto_select_distribute_end_lB_eq_lE(self):
    g = gfapy.Gfa()
    # lB == lE == 1
    self.assertEqual(None, g._auto_select_distribute_end( 4, 1, 1, False))
    # lB == lE == factor
    self.assertEqual("R", g._auto_select_distribute_end( 4, 4, 4, False))
    # lB == lE; </> factor
    self.assertEqual("R", g._auto_select_distribute_end( 4, 2, 2, False))
    self.assertEqual("L", g._auto_select_distribute_end( 4, 6, 6, False))

  def test_auto_select_distribute_end_l_1(self):
    g = gfapy.Gfa()
    # lB or lE == 1, other </==/> factor
    self.assertEqual("L", g._auto_select_distribute_end( 4, 2, 1, False))
    self.assertEqual("L", g._auto_select_distribute_end( 4, 4, 1, False))
    self.assertEqual("L", g._auto_select_distribute_end( 4, 6, 1, False))
    self.assertEqual("R", g._auto_select_distribute_end( 4, 1, 2, False))
    self.assertEqual("R", g._auto_select_distribute_end( 4, 1, 4, False))
    self.assertEqual("R", g._auto_select_distribute_end( 4, 1, 6, False))

  def test_auto_select_distribute_end_eq_factor(self):
    g = gfapy.Gfa()
    # one =, one > factor
    self.assertEqual("L", g._auto_select_distribute_end( 4, 4, 5, False))
    self.assertEqual("R", g._auto_select_distribute_end( 4, 5, 4, False))
    # one =, one < factor
    self.assertEqual("L", g._auto_select_distribute_end( 4, 4, 3, False))
    self.assertEqual("R", g._auto_select_distribute_end( 4, 3, 4, False))

  def test_auto_select_distribute_end_diff_factor(self):
    g = gfapy.Gfa()
    # both > 1; both < factor
    self.assertEqual("L", g._auto_select_distribute_end( 4, 3, 2, False))
    self.assertEqual("R", g._auto_select_distribute_end( 4, 2, 3, False))
    # both > 1; both > factor
    self.assertEqual("L", g._auto_select_distribute_end( 4, 5, 6, False))
    self.assertEqual("R", g._auto_select_distribute_end( 4, 6, 5, False))
    # both > 1; one <, one > factor
    self.assertEqual("L", g._auto_select_distribute_end( 4, 3, 5, False))
    self.assertEqual("R", g._auto_select_distribute_end( 4, 5, 3, False))




import gfapy

class FromTo:

  def is_circular(self):
    """Does the edge represent an alignment of a segment with itself?

    Returns:
      bool : if the sid1/sid2 (E) or from_segment/to_segment (L/C)
             segments are equal to each other; thereby, the orientations
             are not considered
    """
    return str(self.from_segment) == str(self.to_segment)

  def is_circular_same_end(self):
    """Does the edge represent an alignment of a segment end with itself?

    Returns
      bool : if the edge is a dovetail overlap, for which sid1/sid2 (E) or
        from_segment/to_segment (L) are equal to each other, and the same
        end of the sequence overlaps itself; (note that this kind of edge
        is actually quite meaningless, but can be defined)
    """
    return self.from_end == self.to_end

  @property
  def from_end(self):
    """The segment end corresponding to the from_segment field of L lines.

    Note:
      The result is meaningful only for dovetails overlaps (GFA1 L lines
      or GFA2 E lines representing dovetail overlaps).

    For a L line, the from_orient field is used to compute if the overlap
    involves the left (5') or right (3') end of the from_segment and the
    SegmentEnd end_type property is set accordingly to 'L' or 'R'.  For a E
    line, it is first computed which of the sid1/sid2 corresponds to the
    from_segment field of a L line, then the same computation is done, as for L
    lines.

    Returns:
      gfapy.segment_end.SegmentEnd
    """
    return gfapy.SegmentEnd(self.from_segment,
                            "R" if self.from_orient == "+" else "L")

  @property
  def to_end(self):
    """The segment end corresponding to the to_segment field of L lines.

    Note:
      The result is meaningful only for dovetails overlaps (GFA1 L lines
      or GFA2 E lines representing dovetail overlaps).

    For a L line, the to_orient field is used to compute if the overlap
    involves the left (5') or right (3') end of the to_segment and the
    SegmentEnd end_type property is set accordingly to 'L' or 'R'.  For a E
    line, it is first computed which of the sid1/sid2 corresponds to the
    to_segment field of a L line, then the same computation is done, as for L
    lines.

    Returns:
      gfapy.segment_end.SegmentEnd
    """
    return gfapy.SegmentEnd(self.to_segment,
                            "L" if self.to_orient == "+" else "R")

  def other_end(self, segment_end, tolerant=False):
    """The other segment end involved in the alignment represented by the edge.

    Note:
      The result is meaningful only for dovetails overlaps (GFA1 L lines
      or GFA2 E lines representing dovetail overlaps).

    Parameters:
      segment_end (`gfapy.segment_end.SegmentEnd`) : one of the two segment
        ends involved in the alignment represented by the edge

    Returns:
      gfapy.segment_end.SegmentEnd

    Raises:
      gfapy.error.ArgumentError: If segment_end is not a valid segment end
      gfapy.RuntimeError: if the segment_end is not involved in the alignment
        represented by the line.
    """
    segment_end
    if (self.from_end == segment_end):
      return self.to_end
    elif (self.to_end == segment_end):
      return self.from_end
    elif tolerant:
      return None
    else:
      raise gfapy.ArgumentError(
        "Segment end '{}' not found\n".format(repr(segment_end))+
        "(from={};to={})".format(repr(self.from_end), repr(self.to_end)))

  @property
  def from_name(self):
    """Segment name of the segment with the role of a L/C from_segment.

   The method allows to compute the segment name in both cases in which
   the segment is a string (i.e. the segment name itself) or a reference
   to a segment line.

    Returns:
      str : the name of the segment which is specified in the field which
        corresponds to the from_segment field in a GFA1 line (from_segment
        if GFA1, sid1 or sid2 if GFA2)
    """
    if isinstance(self.from_segment, str):
      return self.from_segment
    else:
      return self.from_segment.name

  @property
  def to_name(self):
    """Segment name of the segment with the role of a L/C to_segment.

   The method allows to compute the segment name in both cases in which
   the segment is a string (i.e. the segment name itself) or a reference
   to a segment line.

    Returns:
      str : the name of the segment which is specified in the field which
        corresponds to the to_segment field in a GFA1 line (to_segment
        if GFA1, sid1 or sid2 if GFA2)
    """
    if isinstance(self.to_segment, str):
      return self.to_segment
    else:
      return self.to_segment.name

  @property
  def _segment_ends_s(self):
    """Signature of the segment ends, for debugging."""
    return "---".join([str(self.from_end), str(self.to_end)])



from .common.construction import Construction
from .common.dynamic_fields import DynamicFields, DynamicField
from .common.writer import Writer
from .common.version_conversion import VersionConversion
from .common.field_datatype import FieldDatatype
from .common.field_data import FieldData
from .common.equivalence import Equivalence
from .common.cloning import Cloning
from .common.connection import Connection
from .common.virtual_to_real import VirtualToReal
from .common.update_references import UpdateReferences
from .common.disconnection import Disconnection
from .common.validate import Validate
from .common.default_record_definition import DefaultRecordDefinition

import gfapy

class Line(Construction, DynamicFields, Writer, VersionConversion,
           FieldDatatype, FieldData, Equivalence, Cloning, Connection,
           VirtualToReal, UpdateReferences, Disconnection, Validate,
           DefaultRecordDefinition):
  """
  A line of a GFA file.

  Parameters:
    data (str, list of str) : the content of a line in a GFA file, either as
      a string, or as a list derived from tab-splitting the line string
    vlevel (int) : an integer from 0 to 3, which specifies the validation level;
      if 0, no validation is performed (the user can still validate manually if
      needed); if 1 (the default), validation is performed when the line
      is constructed, or, for some fields, when the value is accessed
      for the first time; if 2, the validation is performed also when converting
      the content of a field to string; if 3, also each time the value
      of a field is read or written
    version (str) : one of 'gfa1' and 'gfa2'; the GFA version; if not specified,
      then the version is guessed from the record type and syntax, or set
      to 'generic'
    dialect (str) : one of 'rgfa' and 'standard'; the GFA dialect; if not
      specified then the dialect is set to 'standard'; 'rgfa' is only valid
      when version is 'gfa1'

  Notes:
    The private interface to the Line constructor also allows to pass a
    dictionary instead of a list for data. Furthermore the private parameter
    virtual allows to create virtual line instances, which are useful during
    parsing.

  Raises:
    gfapy.error.FormatError: If the line contains a wrong number of positional
      fields, or if the content of a field has a wrong format.
    gfapy.error.NotUniqueError: If a tag name is used more than once.
    gfapy.error.TypeError: If the value of a predefined tag does not
      respect the datatype specified in the tag.

  Returns:
    an instance of a subclass of gfapy.line.Line
  """

  SEPARATOR = "\t"
  """Separator in the string representation of GFA lines"""




import unittest
import gfapy

class TestLineHeader(unittest.TestCase):

  def test_from_string(self):
    gfapy.Line("H\tVN:Z:1.0")
    self.assertIsInstance(gfapy.Line("H\tVN:Z:1.0"), gfapy.line.Header)
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line("H\tH2\tVN:Z:1.0")
    with self.assertRaises(gfapy.TypeError):
      gfapy.Line("H\tVN:i:1.0")



import gfapy
from .artifacts import Artifacts
from .copy_number import CopyNumber
from .invertible_segments import InvertibleSegments
from .p_bubbles import PBubbles
from .linear_paths import LinearPaths
from .multiplication import Multiplication
from .redundant_linear_paths import RedundantLinearPaths
from .superfluous_links import SuperfluousLinks
from .topology import Topology
class GraphOperations(LinearPaths,Multiplication,RedundantLinearPaths,
    Topology,Artifacts,CopyNumber,InvertibleSegments,PBubbles,SuperfluousLinks):
  pass



import unittest
import gfapy

class TestUnitLineDynamicFields(unittest.TestCase):

  def test_respond_to(self):
    l = gfapy.line.edge.Link(["L", "1", "+", "2", "-", "*", "zz:Z:yes", "KC:i:100"])
    # record_type
    self.assertTrue(hasattr(l, "record_type"))
    # reqfields
    self.assertTrue(hasattr(l, "from"))
    self.assertIsInstance(object.__getattribute__(l, "from"),
                          gfapy.line.common.dynamic_fields.DynamicField)
    # predefined tags
    self.assertTrue(hasattr(l, "KC"))
    self.assertTrue(hasattr(l, "try_get_KC"))
    self.assertIsInstance(object.__getattribute__(l, "KC"),
                          gfapy.line.common.dynamic_fields.DynamicField)
    # custom tags
    self.assertTrue(hasattr(l, "zz"))
    self.assertTrue(hasattr(l, "try_get_zz"))
    # not-yet-existing tags
    self.assertTrue(hasattr(l, "aa"))
    #raises exception in python, because hasattr calls getattr
    #self.assertTrue(hasattr(l, "try_get_aa"))

  def test_field_getters_positional_fields(self):
    l = gfapy.Line(["S", "12", "*", "xx:i:13", "KC:i:10"])
    self.assertEqual("12", l.name)
    with self.assertRaises(AttributeError):
      l.zzz

  def test_field_getters_existing_tags(self):
    l = gfapy.Line(["S", "12", "*", "xx:i:13", "KC:i:10"])
    self.assertEqual("xx", sorted(l.tagnames)[1])
    self.assertEqual("13", l.field_to_s("xx"))
    self.assertEqual(13, l.xx)
    self.assertEqual(13, l.try_get_xx())
    self.assertEqual("10", l.field_to_s("KC"))
    self.assertEqual(10, l.KC)
    self.assertEqual(10, l.try_get_KC())

  def test_field_getters_not_existing_tags(self):
    l = gfapy.line.Header(["H", "xx:i:13", "VN:Z:HI"])
    self.assertEqual(None, l.zz)
    with self.assertRaises(gfapy.NotFoundError):
      l.try_get_zz()

  def test_field_setters_positional_fields(self):
    l = gfapy.Line(["S", "12", "*", "xx:i:13", "KC:i:1200"])
    with self.assertRaises(gfapy.FormatError):
      l.name = "A\t1"
      l.validate_field("name")
    l.name = "14"
    self.assertEqual("14", l.name)

  def test_field_setters_existing_tags(self):
    l = gfapy.line.Header(["H", "xx:i:13", "VN:Z:HI"], vlevel = 3)
    self.assertEqual(13, l.xx)
    l.xx = 15
    self.assertEqual(15, l.xx)
    with self.assertRaises(gfapy.FormatError):
      l.xx = "1A"
    l.set_datatype("xx", "Z")
    l.xx = "1A"
    self.assertEqual("HI", l.VN)
    l.VN = "HO"
    self.assertEqual("HO", l.VN)

  def test_field_setters_not_existing_tags(self):
    l = gfapy.line.Header(["H", "xx:i:13", "VN:Z:HI"])
    l.zz="1"
    self.assertEqual("1", l.zz)
    self.assertEqual("Z", gfapy.Field._get_default_gfa_tag_datatype(l.zz))
    l.zi=1
    self.assertEqual(1, l.zi)
    self.assertEqual("i", gfapy.Field._get_default_gfa_tag_datatype(l.zi))
    l.zf=1.0
    self.assertEqual(1.0, l.zf)
    self.assertEqual("f", gfapy.Field._get_default_gfa_tag_datatype(l.zf))
    l.bf=[1.0, 1.0]
    self.assertEqual([1.0, 1.0], l.bf)
    self.assertEqual("B", gfapy.Field._get_default_gfa_tag_datatype(l.bf))
    l.bi=[1.0, 1.0]
    self.assertEqual([1, 1], l.bi)
    self.assertEqual("B", gfapy.Field._get_default_gfa_tag_datatype(l.bi))
    l.ba=[1.0, 1]
    self.assertEqual([1.0, 1], l.ba)
    self.assertEqual("J", gfapy.Field._get_default_gfa_tag_datatype(l.ba))
    l.bh={"a" : 1.0, "b" : 1}
    self.assertEqual({"a" : 1.0, "b" : 1}, gfapy.Line(str(l)).bh)
    self.assertEqual("J", gfapy.Field._get_default_gfa_tag_datatype(l.bh))
    #Assignement of new attributes possible in python.
    #with self.assertRaises(AttributeError):
    #  l.zzz="1"




import gfapy

class Complement:

  def complement(self):
    """Creates the equivalent link with from and to inverted.

    The CIGAR operations (order and type) are inverted as well.
    Tags are left unchanged.

    Note:
        The path references are not copied to the complement link.

    Note:
        This method shall be overridden if custom tags are defined, which have a
        complementation operation which determines their value in the
        equivalent complement link.

    Returns:
        gfapy.line.edge.Link: The inverted link.
    """
    l = self.clone()
    l.from_segment = self.to
    l.from_orient = gfapy.invert(self.to_orient)
    l.to_segment = self.from_segment
    l.to_orient = gfapy.invert(self.from_orient)
    l.overlap = self.overlap.complement()
    return l

  def make_complement(self):
    """Complements the link inplace.

    The tags are left unchanged.

    Note:
        The path references are not complemented by this method; therefore
        the method shall be used before the link is embedded in a graph.

    Note:
        This method shall be overridden if custom tags are defined, which have a
        complementation operation which determines their value in the
        complement link.

    Returns:
        gfapy.line.edge.Link: self
    """
    tmp = self.from_segment
    self.from_segment = self.to_segment
    self.to_segment = tmp
    tmp = self.from_orient
    self.from_orient = gfapy.invert(self.to_orient)
    self.to_orient = gfapy.invert(tmp)
    self.overlap = self.overlap.complement()
    return self



import gfapy

class References:
  def _initialize_references(self):
    for d in ["from", "to"]:
      s = self._gfa.segment(self.get(d))
      if s is None:
        if self._gfa._segments_first_order:
          raise gfapy.NotFoundError()
        s = gfapy.line.segment.GFA1({"name" : self.get(d),
                                     "sequence" : "*"},
                                    version = "gfa1",
                                    virtual = True)
        s.connect(self._gfa)
      self._set_existing_field(d+"_segment", s, set_reference = True)
      if self.record_type == "L":
        et = self.from_end.end_type if d == "from" else self.to_end.end_type
        key = "dovetails_{}".format(et)
      else:
        key = \
          "edges_to_contained" if (d == "from") else "edges_to_containers"
      s._add_reference(self, key)

  def _import_field_references(self, previous):
    for d in ["from_segment", "to_segment"]:
      self._set_existing_field(d, self._gfa.segment(self.get(d)),
          set_reference = True)

  def _backreference_keys(self, ref, key_in_ref):
    if ref.record_type == "P":
      return ["paths"]
    elif ref.record_type == "S":
      return ["from_segment", "to_segment"]
    else:
      return []



import gfapy
from .references import References
from ..line import Line

class Gap(References, Line):
  """
  A gap line of a GFA2 file
  """
  RECORD_TYPE = "G"
  POSFIELDS = ["gid", "sid1", "sid2", "disp", "var"]
  FIELD_ALIAS = { "name" : "gid" }
  NAME_FIELD = "gid"
  STORAGE_KEY = "name"
  DATATYPE = {
    "gid" : "optional_identifier_gfa2",
    "sid1" : "oriented_identifier_gfa2",
    "sid2" : "oriented_identifier_gfa2",
    "disp" : "i",
    "var" : "optional_integer"
  }
  REFERENCE_FIELDS = ["sid1", "sid2"]

Gap._apply_definitions()



"""
Decoding of the GFA string representations into python objects
"""
import gfapy
import re

class Parser:

  @staticmethod
  def _parse_gfa_field(string, datatype, safe = True, fieldname = None,
                      line = None):
    """
    Parse a GFA string representation and decodes it into a python object

    Parameters
    ----------
    string : str
      the GFA string to parse
    datatype : one of gfapy.Field.FIELD_DATATYPE
      the datatype to use
    safe : bool, optional
      *(defaults to: ***True***)* if **True** the safe
      version of the decode function for the datatype is used, which
      validates the content of the string; if **False**, the string is
      assumed to be valid and decoded into a value accordingly, which may
      result in invalid values (but may be faster than the safe decoding)
    fieldname : str, optional
      fieldname for error messages
    line : gfapy.Line, optional
      line content for error messages

    Raises
    ------
    gfapy.TypeError
      if the specified datatype is unknown
    gfapy.FormatError
      if the string syntax is not valid
    gfapy.ValueError
      if the decoded value is not valid
    """
    mod = gfapy.Field.FIELD_MODULE.get(datatype)
    if mod is None:
      linemsg = ""
      try:
        if line is not None and not line.__error__:
          line.__error__ = True # avoids infinite recursion
          linemsg = ["Line content:"]
          linemsg.append(str(line))
          linemsg.append("\n")
      except:
        pass
      fieldnamemsg = "Field: {}\n".format(fieldname) if fieldname else ""
      contentmsg = "Content: {}\n".format(string)
      raise gfapy.TypeError(
        linemsg +
        fieldnamemsg +
        contentmsg +
        "Datatype unknown: {}".format(repr(datatype)))
    try:
      if safe or not getattr(mod, "unsafe_decode"):
        return mod.decode(string)
      else:
        return mod.unsafe_decode(string)
    except Exception as err:
      linemsg = ""
      try:
        if line is not None and not line.__error__:
          line.__error__ = True # avoids infinite recursion
          linemsg = ["Line content:"]
          linemsg.append(str(line))
          linemsg.append("\n")
      except:
        pass
      fieldnamemsg = "Field: {}\n".format(fieldname) if fieldname else ""
      contentmsg = "Content: {}\n".format(string)
      datatypemsg = "Datatype: {}\n".format(datatype)
      errmsg = err.message if hasattr(err, "message") else str(err)
      raise err.__class__(
            linemsg +
            fieldnamemsg +
            datatypemsg +
            contentmsg +
            errmsg) from err

  @staticmethod
  def _parse_gfa_tag(tag):
    """
    Parses a GFA tag in the form **xx:d:content** into its components.
    The **content** is not decoded (see :func:`_parse_gfa_field`).

    Parameters
    ----------
    tag : str
      the GFA tag to parse

    Raises
    ------
    gfapy.FormatError
      if the string does not represent a valid GFA tag

    Returns
    -------
    list of (str, gfapy.Field.FIELD_DATATYPE)
      the parsed content of the field
    """
    match = re.match(r"^([A-Za-z][A-Za-z0-9]):([AifZJHB]):(.+)$", tag)
    if match:
      return [match.group(1), match.group(2), match.group(3)]
    else:
      raise gfapy.FormatError(
        "Expected GFA tag, found: {}".format(repr(tag)))



import unittest
import gfapy

class TestLineContainment(unittest.TestCase):

  def test_from_string(self):
    fields = ["C","1","+","2","-","12","12M","MQ:i:1232","NM:i:3","ab:Z:abcd"]
    string="\t".join(fields)
    gfapy.Line(string)
    self.assertIsInstance(gfapy.Line(string), gfapy.line.edge.Containment)
    self.assertEqual(fields[0], gfapy.Line(string).record_type)
    self.assertEqual(fields[1], gfapy.Line(string).from_segment)
    self.assertEqual(fields[2], gfapy.Line(string).from_orient)
    self.assertEqual(fields[3], gfapy.Line(string).to_segment)
    self.assertEqual(fields[4], gfapy.Line(string).to_orient)
    self.assertEqual(12, gfapy.Line(string).pos)
    self.assertEqual([gfapy.alignment.cigar.CIGAR.Operation(12, "M")],
                     gfapy.Line(string).overlap)
    self.assertEqual(1232, gfapy.Line(string).MQ)
    self.assertEqual(3, gfapy.Line(string).NM)
    self.assertEqual("abcd", gfapy.Line(string).ab)
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line(string+"\tH1")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line(string+"\tH1")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line("C\tH")
    with self.assertRaises(gfapy.FormatError):
      f=fields[:]
      f[2]="x"
      gfapy.Line("\t".join(f), vlevel = 2)
    with self.assertRaises(gfapy.FormatError):
      f=fields[:]
      f[4]="x"
      gfapy.Line("\t".join(f), vlevel = 2)
    with self.assertRaises(gfapy.FormatError):
      f=fields[:]
      f[5]="x"
      gfapy.Line("\t".join(f), vlevel = 2)
    with self.assertRaises(gfapy.FormatError):
      f=fields[:]
      f[6]="x"
      gfapy.Line("\t".join(f), vlevel = 2)
    with self.assertRaises(gfapy.TypeError):
      f=fields[:]
      f[7]="MQ:Z:1232"
      gfapy.Line("\t".join(f), vlevel = 2)
    with self.assertRaises(gfapy.TypeError):
      f=fields[:]
      f[8]="NM:Z:1232"
      gfapy.Line("\t".join(f), vlevel = 2)



import gfapy

class Destructors:

  def rm(self, gfa_line):
    """Remove a line from the Gfa instance.

    Parameters:
      gfa_line (Line, str): if a Line instance, then disconnect is called
        on it; otherwise, it is assumed to be a line identifier, the
        corresponding line is searched and then disconnected.

    Raises:
      gfapy.error.NotFoundError : is the line is specified using an identifier,
        an no line exists in the Gfa instance, with that identifier
    """
    self.try_get_line(gfa_line).disconnect()

  def _delete_other_links(self, segment_end, other_end,
                         conserve_components = False):
    segment_end = gfapy.SegmentEnd(segment_end)
    other_end = gfapy.SegmentEnd(other_end)
    s = self.try_get_segment(segment_end.segment)
    for d in s.dovetails_of_end(segment_end.end_type):
      if not conserve_components or not self.is_cut_link(l):
        l.disconnect()

  def _unregister_line(self, gfa_line):
    self._api_private_check_gfa_line(gfa_line, "unregister_line")
    rt = gfa_line.record_type
    if rt == "H":
      raise gfapy.AssertionError("Bug found, please report\n"+
        "gfa_line: {}".format(gfa_line))
    collection = self._records[rt]
    key = gfa_line
    delete_if_empty = None
    storage_key = gfa_line.__class__.STORAGE_KEY
    if storage_key == "name":
      name = gfa_line.name
      if gfapy.is_placeholder(name):
        name = id(gfa_line)
      collection.pop(name)
    elif storage_key == "external":
      subkey = gfa_line.external.name
      collection = collection[subkey]
      collection.pop(id(gfa_line))
      if not collection:
        self._records[rt].pop(subkey)
    else:
      collection.pop(id(gfa_line))



import gfapy

class Connection:

  def connect(self, gfa):
    if gfa.header is not self:
      raise gfapy.RuntimeError(
        "gfapy.line.Header instances cannot be connected\n"+
        "Use gfa.add_line(this_line) to add the information\n"+
        "contained in this header line to the header of a GFA instance.")
    else:
      self._gfa = gfa



import gfapy

class Multiline:
  """
  Implementation of the support for multiple header line in the
  GFA file (which also may contain the same value defined multiple
  times in different lines).
  """

  SINGLE_DEFINITION_TAGS = ["VN", "TS"]

  def add(self, tagname, value, datatype = None):
    """
    Set a header value (multi-value compatible).

    If a field does not exist yet, set it to value. If it exists and it is a
    *gfapy.FieldArray*, add the value to the field array. If it exists and it
    is not a field array, create a field array with the previous value and
    the new one.

    Parameters
    ----------
    tagname : str
    value : object
    datatype : gfapy.Field.TAG_DATATYPE, optional
      The datatype to use.
      The default is to determine the datatype according to the value or the
      previous values present in the field.
    """
    prev = self.get(tagname)
    if prev is None:
      if datatype is not None:
        self.set_datatype(tagname, datatype)
      self.set(tagname, value)
      return
    elif not isinstance(prev, gfapy.FieldArray):
      if tagname in self.SINGLE_DEFINITION_TAGS:
        if self.field_to_s(tagname) == \
            gfapy.Field._to_gfa_field(value, fieldname=tagname):
          return
        else:
          raise gfapy.InconsistencyError(
            "Inconsistent values for header tag {} found\n".format(tagname)+
            "Previous definition: {}\n".format(prev)+
            "Current definition: {}".format(value))
      prev = gfapy.FieldArray(self.get_datatype(tagname), [prev])
      self._set_existing_field(tagname, prev)
    if self.vlevel > 1:
      prev.vpush(value, datatype, tagname)
    else:
      prev.append(value)

  def field_to_s(self, fieldname, tag = False):
    """
    Compute the string representation of a field.

    Parameters
    ----------
    fieldname : str
      The tag name of the field.
    tag : bool
      *(defaults to: ***False***)*
      Return the 'tagname:datatype:value' representation.

    Raises
    ------
    gfapy.NotFoundError
      If field is not defined.

    Returns
    -------
    str
      The string representation.
    """
    prev = self.get(fieldname)
    if isinstance(prev, gfapy.FieldArray):
      if self.vlevel >= 2:
        prev._validate_gfa_field(None, fieldname)
      return prev._to_gfa_tag(fieldname=fieldname) if tag else \
             prev._to_gfa_field(fieldname=fieldname)
    else:
      return super(gfapy.line.header.Line, self).field_to_s(fieldname, tag)

  def _n_duptags(self):
    n = 0
    for tn in self.tagnames:
      if isinstance(self.get(tn),gfapy.FieldArray):
        n+=1
    return n

  def _split(self):
    """
    Split the header line into single-tag lines.

    If a tag is a FieldArray, this is splitted into multiple fields
    with the same fieldname.

    Returns
    -------
    gfapy.line.Header list
    """
    retval = []
    for tagname, datatype, value in self._tags():
      h = gfapy.line.Header(["H"], vlevel = self.vlevel)
      h.set_datatype(tagname, datatype)
      h.set(tagname, value)
      retval.append(h)
    return retval

  def _merge(self, gfa_line):
    """
    Merge an additional **gfa.line.Header** line into this header line.

    Parameters
    ----------
    gfa_line : gfapy.line.Header
      The header line to merge.

    Returns
    -------
    self
    """
    for of in gfa_line.tagnames:
      self.add(of, gfa_line.get(of), gfa_line.get_datatype(of))
    return self

  def _tags(self):
    """
    List of tags data.

    Returns the tags as an list of [fieldname, datatype, value]
    lists. If a field is a FieldArray, this is splitted into multiple fields
    with the same fieldname.

    Returns
    -------
    (str, str, object) list
    """
    retval = []
    for of in self.tagnames:
      value = self.get(of)
      if isinstance(value, gfapy.FieldArray):
        for elem in value:
          retval.append((of, value.datatype, elem))
      else:
        retval.append((of, self.get_datatype(of), value))
    return retval



"""
Methods in this module are important for lines
which can be virtual
"""
import gfapy

class VirtualToReal:

  @property
  def virtual(self):
    """
    Is the line virtual?

    Is this gfapy.Line a virtual line representation
    (i.e. a placeholder for an expected but not yet encountered line)?

    Returns
    -------
    bool
    """
    return self._virtual

  def _substitute_virtual_line(self, previous):
    self._gfa = previous.gfa
    self._import_references(previous)
    self._gfa._unregister_line(previous)
    self._gfa._register_line(self)
    return None

  def _import_references(self, previous):
    """
    This is called when a virtual line (previous) is
    substituted by a real line
    """
    if not isinstance(previous, gfapy.line.Unknown):
      self._import_field_references(previous)
      self._update_field_backreferences(previous)
    else:
      self._initialize_references()
    self._import_nonfield_references(previous)
    self._update_nonfield_backreferences(previous)

  def _import_field_references(self, previous):
    for k in (self.__class__.REFERENCE_FIELDS +
              self.__class__.BACKREFERENCE_RELATED_FIELDS):
      ref = previous.get(k)
      self._set_existing_field(k, ref, set_reference = True)

  def _update_backreference_in(self, ref, previous, k):
    if isinstance(ref, gfapy.Line):
      ref._update_references(previous, self, k)
    elif isinstance(ref, gfapy.OrientedLine):
      ref.line._update_references(previous, self, k)
    elif isinstance(ref, list):
      for item in ref:
        self._update_backreference_in(item, previous, k)

  def _update_field_backreferences(self, previous):
    """
    .. note::
      Currently this method supports fields which are: references,
      oriented lines and lists of references of oriented lines.
      If SUBCLASSES have reference fields which contain references
      in a different fashion, the method must be updated or overwritten
      in the subclass.
    """
    for k in self.__class__.REFERENCE_FIELDS:
      ref = self.get(k)
      self._update_backreference_in(ref, previous, k)

  def _import_nonfield_references(self, previous):
    self._refs = previous._refs

  def _update_nonfield_backreferences(self, previous):
    for k, v in self._refs.items():
      for ref in v:
        self._update_backreference_in(ref, previous, k)



import gfapy

class Other:

  def other_oriented_segment(self, oriented_segment):
    """The other oriented segment.

    Parameters:
      oriented_segment (gfapy.OrientedLine) : One of the two oriented segments of the line.

    Returns:
      gfapy.OrientedLine

    Raises:
      gfapy.error.NotFoundError: If segment_end is not a segment end of the line.
    """
    if (self.sid1 == oriented_segment):
      return self.sid2
    elif (self.sid2 == oriented_segment):
      return self.sid1
    else:
      raise gfapy.NotFoundError(
          "Oriented segment '{}' not found\n".format(oriented_segment) +
          "Line: {}".format(self))

  def other(self, segment):
    """The other segment of an edge line.

    Parameters:
      segment (gfapy.line.segment.GFA2) : Segment name or instance.

    Raises:
      gfapy.error.NotFoundError: If segment is not a segment of the line.

    Returns:
      gfapy.line.segment.GFA2 or str : instance or name of the other segment
        of the connection (the segment itself, if the connection is circular)
    """
    if isinstance(segment, gfapy.Line):
      segment = segment.name
    if segment == self.sid1.name:
      return self.sid2.line
    elif segment == self.sid2.name:
      return self.sid1.line
    else:
      raise gfapy.NotFoundError(
        "Line {} does not involve segment {}".format(self, segment))



import gfapy
import unittest

class TestLinePath(unittest.TestCase):

  def test_from_string(self):
    fields = ["P","4","1+,2-,3+","9M2I3D1M,12M","ab:Z:abcd"]
    string = "\t".join(fields)
    gfapy.Line(string)
    self.assertIsInstance(gfapy.Line(string), gfapy.line.group.Path)
    self.assertEqual(str(fields[0]), gfapy.Line(string).record_type)
    self.assertEqual(str(fields[1]), gfapy.Line(string).path_name)
    self.assertEqual([gfapy.OrientedLine("1","+"),
                      gfapy.OrientedLine("2","-"),
                      gfapy.OrientedLine("3","+")],
                     gfapy.Line(string).segment_names)
    self.assertEqual([[gfapy.alignment.cigar.Operation(9,"M"),
                       gfapy.alignment.cigar.Operation(2,"I"),
                       gfapy.alignment.cigar.Operation(3,"D"),
                       gfapy.alignment.cigar.Operation(1,"M")],
                      [gfapy.alignment.cigar.Operation(12,"M")]],
                     gfapy.Line(string).overlaps)
    self.assertEqual("abcd", gfapy.Line(string).ab)
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line(string+"\tH1")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line("P\tH")
    with self.assertRaises(gfapy.FormatError):
      f=fields[:]
      f[2]="1,2,3"
      gfapy.Line("\t".join(f), vlevel = 2)
    with self.assertRaises(gfapy.InconsistencyError):
      f=fields[:]
      f[2]="1+,2+"
      f[3]="9M,12M,3M"
      gfapy.Line("\t".join(f), vlevel = 2)

    f=fields[:]
    f[3]="*,*"
    gfapy.Line("\t".join(f), vlevel = 2)

    f=fields[:]
    f[3]="9M2I3D1M,12M,12M"
    gfapy.Line("\t".join(f), vlevel = 2)

    f=fields[:]
    f[3]="*"
    gfapy.Line("\t".join(f), vlevel = 2)

    with self.assertRaises(gfapy.FormatError):
      f=fields[:]
      f[3]="12,12"
      gfapy.Line("\t".join(f), vlevel = 2)
    with self.assertRaises(gfapy.FormatError):
      f=fields[:]
      f[3]="12M|12M"
      gfapy.Line("\t".join(f), vlevel = 2)



class ToGFA2:

  @property
  def from_coords(self):
    """
    GFA2 positions of the alignment on the **from** segment.

    Returns
    -------
    (Integer|Lastpos, Integer|Lastpos)
    	begin and end

    Raises
    ------
    gfapy.RuntimeError
    	If the segment length cannot be determined, because the segment line is unknown.
    gfapy.ValueError
    	If the segment length is not specified in the segment line.
    """
    self._check_overlap()
    rpos = self.pos + self.overlap.length_on_reference()
    if rpos == self._lastpos_of("from"):
      rpos = gfapy.LastPos(rpos)
    return [self.pos, rpos]

  @property
  def to_coords(self):
    """
    GFA2 positions of the alignment on the **to** segment
    """
    return [0, self._lastpos_of("to")]



import gfapy
import unittest

class TestAPIrGfa(unittest.TestCase):

  def test_adding_invalid_segment_to_rgfa(self):
    gfa = gfapy.Gfa()
    gfa.append("S\t1\t*")
    gfa.validate()
    gfa = gfapy.Gfa(dialect="rgfa")
    gfa.append("S\t1\t*")
    with self.assertRaises(gfapy.NotFoundError): gfa.validate()

  def test_adding_containment_to_rgfa(self):
    gfa = gfapy.Gfa()
    gfa.append("C\t1\t+\t2\t+\t12\t*")
    gfa.validate()
    gfa = gfapy.Gfa(version="gfa1",dialect="rgfa")
    gfa.append("C\t1\t+\t2\t+\t12\t*")
    with self.assertRaises(gfapy.NotFoundError): gfa.validate()

  def test_loading_examples(self):
    gfapy.Gfa.from_file("tests/testdata/rgfa_example.1.gfa", dialect="rgfa")
    gfapy.Gfa.from_file("tests/testdata/rgfa_example.2.gfa", dialect="rgfa")

  def test_stable_sequence_names(self):
    g = gfapy.Gfa.from_file("tests/testdata/rgfa_example.2.gfa", dialect="rgfa")
    self.assertEqual(['smpl-Ref.Bd4', 'smpl-Bd21_3_r.pseudomolecule_4'],
        g.stable_sequence_names)
    g = gfapy.Gfa.from_file("tests/testdata/rgfa_example.1.gfa", dialect="rgfa")
    self.assertEqual(['bar', 'foo', 'chr1'],
        g.stable_sequence_names)
